"""This file contains factory methods which return template prompts and return messages for chat payloads."""

import base64
import json
from string import Formatter

from palimpzest.constants import (
    MIXTRAL_LLAMA_CONTEXT_TOKENS_LIMIT,
    TOKENS_PER_CHARACTER,
    Cardinality,
    Model,
    PromptStrategy,
)
from palimpzest.core.elements.records import DataRecord
from palimpzest.core.lib.fields import BytesField, ImageBase64Field, ImageFilepathField, ImageURLField
from palimpzest.core.lib.schemas import Schema
from palimpzest.prompts.convert_prompts import (
    COT_QA_BASE_SYSTEM_PROMPT,
    COT_QA_BASE_USER_PROMPT,
    COT_QA_EXAMPLE_ANSWER,
    COT_QA_EXAMPLE_CONTEXT,
    COT_QA_EXAMPLE_INPUT_FIELDS,
    COT_QA_EXAMPLE_OUTPUT_FIELDS,
    COT_QA_EXAMPLE_REASONING,
    COT_QA_IMAGE_DISCLAIMER,
    COT_QA_IMAGE_EXAMPLE_ANSWER,
    COT_QA_IMAGE_EXAMPLE_CONTEXT,
    COT_QA_IMAGE_EXAMPLE_INPUT_FIELDS,
    COT_QA_IMAGE_EXAMPLE_OUTPUT_FIELDS,
    COT_QA_IMAGE_EXAMPLE_REASONING,
    COT_QA_IMAGE_JOB_INSTRUCTION,
    COT_QA_JOB_INSTRUCTION,
)
from palimpzest.prompts.critique_and_refine_convert_prompts import (
    BASE_CRITIQUE_PROMPT,
    BASE_REFINEMENT_PROMPT,
    COT_QA_CRITIQUE_CRITERIA,
    COT_QA_CRITIQUE_FINISH_INSTRUCTION,
    COT_QA_IMAGE_CRITIQUE_CRITERIA,
    COT_QA_IMAGE_REFINEMENT_CRITERIA,
    COT_QA_REFINEMENT_CRITERIA,
    COT_QA_REFINEMENT_FINISH_INSTRUCTION,
)
from palimpzest.prompts.filter_prompts import (
    COT_BOOL_BASE_SYSTEM_PROMPT,
    COT_BOOL_BASE_USER_PROMPT,
    COT_BOOL_EXAMPLE_CONTEXT,
    COT_BOOL_EXAMPLE_FILTER_CONDITION,
    COT_BOOL_EXAMPLE_INPUT_FIELDS,
    COT_BOOL_EXAMPLE_REASONING,
    COT_BOOL_IMAGE_DISCLAIMER,
    COT_BOOL_IMAGE_EXAMPLE_CONTEXT,
    COT_BOOL_IMAGE_EXAMPLE_FILTER_CONDITION,
    COT_BOOL_IMAGE_EXAMPLE_INPUT_FIELDS,
    COT_BOOL_IMAGE_EXAMPLE_REASONING,
    COT_BOOL_IMAGE_JOB_INSTRUCTION,
    COT_BOOL_JOB_INSTRUCTION,
)
from palimpzest.prompts.moa_aggregator_convert_prompts import (
    COT_MOA_AGG_BASE_SYSTEM_PROMPT,
    COT_MOA_AGG_BASE_USER_PROMPT,
)
from palimpzest.prompts.moa_proposer_convert_prompts import (
    COT_MOA_PROPOSER_BASE_SYSTEM_PROMPT,
    COT_MOA_PROPOSER_BASE_USER_PROMPT,
    COT_MOA_PROPOSER_EXAMPLE_ANSWER,
    COT_MOA_PROPOSER_EXAMPLE_CONTEXT,
    COT_MOA_PROPOSER_EXAMPLE_INPUT_FIELDS,
    COT_MOA_PROPOSER_EXAMPLE_OUTPUT_FIELDS,
    COT_MOA_PROPOSER_IMAGE_DISCLAIMER,
    COT_MOA_PROPOSER_IMAGE_EXAMPLE_ANSWER,
    COT_MOA_PROPOSER_IMAGE_EXAMPLE_CONTEXT,
    COT_MOA_PROPOSER_IMAGE_EXAMPLE_INPUT_FIELDS,
    COT_MOA_PROPOSER_IMAGE_EXAMPLE_OUTPUT_FIELDS,
    COT_MOA_PROPOSER_IMAGE_JOB_INSTRUCTION,
    COT_MOA_PROPOSER_JOB_INSTRUCTION,
)
from palimpzest.prompts.split_merge_prompts import (
    COT_SPLIT_MERGER_BASE_SYSTEM_PROMPT,
    COT_SPLIT_MERGER_BASE_USER_PROMPT,
)
from palimpzest.prompts.split_proposer_prompts import (
    COT_SPLIT_PROPOSER_BASE_SYSTEM_PROMPT,
    COT_SPLIT_PROPOSER_BASE_USER_PROMPT,
    SPLIT_PROPOSER_EXAMPLE_ANSWER,
    SPLIT_PROPOSER_EXAMPLE_CONTEXT,
    SPLIT_PROPOSER_EXAMPLE_INPUT_FIELDS,
    SPLIT_PROPOSER_EXAMPLE_OUTPUT_FIELDS,
    SPLIT_PROPOSER_JOB_INSTRUCTION,
)
from palimpzest.prompts.util_phrases import (
    ONE_TO_MANY_OUTPUT_FORMAT_INSTRUCTION,
    ONE_TO_ONE_OUTPUT_FORMAT_INSTRUCTION,
)


class PromptFactory:
    """Factory class for generating prompts for the Generator given the input(s)."""

    BASE_SYSTEM_PROMPT_MAP = {
        PromptStrategy.COT_BOOL: COT_BOOL_BASE_SYSTEM_PROMPT,
        PromptStrategy.COT_BOOL_IMAGE: COT_BOOL_BASE_SYSTEM_PROMPT,
        PromptStrategy.COT_QA: COT_QA_BASE_SYSTEM_PROMPT,
        PromptStrategy.COT_QA_CRITIC: None,
        PromptStrategy.COT_QA_REFINE: None,
        PromptStrategy.COT_QA_IMAGE: COT_QA_BASE_SYSTEM_PROMPT,
        PromptStrategy.COT_QA_IMAGE_CRITIC: None,
        PromptStrategy.COT_QA_IMAGE_REFINE: None,
        PromptStrategy.COT_MOA_PROPOSER: COT_MOA_PROPOSER_BASE_SYSTEM_PROMPT,
        PromptStrategy.COT_MOA_PROPOSER_IMAGE: COT_MOA_PROPOSER_BASE_SYSTEM_PROMPT,
        PromptStrategy.COT_MOA_AGG: COT_MOA_AGG_BASE_SYSTEM_PROMPT,
        PromptStrategy.SPLIT_PROPOSER: COT_SPLIT_PROPOSER_BASE_SYSTEM_PROMPT,
        PromptStrategy.SPLIT_MERGER: COT_SPLIT_MERGER_BASE_SYSTEM_PROMPT,
    }
    BASE_USER_PROMPT_MAP = {
        PromptStrategy.COT_BOOL: COT_BOOL_BASE_USER_PROMPT,
        PromptStrategy.COT_BOOL_IMAGE: COT_BOOL_BASE_USER_PROMPT,
        PromptStrategy.COT_QA: COT_QA_BASE_USER_PROMPT,
        PromptStrategy.COT_QA_CRITIC: BASE_CRITIQUE_PROMPT,
        PromptStrategy.COT_QA_REFINE: BASE_REFINEMENT_PROMPT,
        PromptStrategy.COT_QA_IMAGE: COT_QA_BASE_USER_PROMPT,
        PromptStrategy.COT_QA_IMAGE_CRITIC: BASE_CRITIQUE_PROMPT,
        PromptStrategy.COT_QA_IMAGE_REFINE: BASE_REFINEMENT_PROMPT,
        PromptStrategy.COT_MOA_PROPOSER: COT_MOA_PROPOSER_BASE_USER_PROMPT,
        PromptStrategy.COT_MOA_PROPOSER_IMAGE: COT_MOA_PROPOSER_BASE_USER_PROMPT,
        PromptStrategy.COT_MOA_AGG: COT_MOA_AGG_BASE_USER_PROMPT,
        PromptStrategy.SPLIT_PROPOSER: COT_SPLIT_PROPOSER_BASE_USER_PROMPT,
        PromptStrategy.SPLIT_MERGER: COT_SPLIT_MERGER_BASE_USER_PROMPT,
    }

    def __init__(self, prompt_strategy: PromptStrategy, model: Model, cardinality: Cardinality) -> None:
        self.prompt_strategy = prompt_strategy
        self.model = model
        self.cardinality = cardinality

    def _get_context(self, candidate: DataRecord, input_fields: list[str]) -> str:
        """
        Returns the context for the prompt.

        Args:
            candidate (DataRecord): The input record.
            input_fields (list[str]): The input fields.

        Returns:
            str: The context.
        """
        # get context from input record (project_cols will be None if not provided in kwargs)
        context: dict = candidate.to_dict(include_bytes=False, project_cols=input_fields)

        # TODO: MOVE THIS LOGIC INTO A CHUNKING / CONTEXT MANAGEMENT CLASS
        #   - this class should be able to:
        #      - handle the context length of different models (i.e. self.model should be an input)
        #      - handle images
        #      - handle the issue with `original_messages` (ask Matt if this is not clear)
        # TODO: this does not work for image prompts
        # TODO: this ignores the size of the `orignal_messages` in critique and refine prompts
        # cut down on context based on window length
        if self.model.is_llama_model() or self.model.is_mixtral_model():
            total_context_len = len(json.dumps(context, indent=2))

            # sort fields by length and progressively strip from the longest field until it is short enough;
            # NOTE: MIXTRAL_LLAMA_CONTEXT_TOKENS_LIMIT is a rough estimate which leaves room for the rest of the prompt text
            while total_context_len * TOKENS_PER_CHARACTER > MIXTRAL_LLAMA_CONTEXT_TOKENS_LIMIT:
                # sort fields by length
                field_lengths = [(field, len(value) if value is not None else 0) for field, value in context.items()]
                sorted_fields = sorted(field_lengths, key=lambda item: item[1], reverse=True)

                # get field with longest context
                longest_field_name, longest_field_length = sorted_fields[0]

                # trim the field
                context_factor = MIXTRAL_LLAMA_CONTEXT_TOKENS_LIMIT / (total_context_len * TOKENS_PER_CHARACTER)
                keep_frac_idx = int(longest_field_length * context_factor)
                context[longest_field_name] = context[longest_field_name][:keep_frac_idx]

                # update total context length
                total_context_len = len(json.dumps(context, indent=2))

        return json.dumps(context, indent=2)

    def _get_input_fields(self, candidate: DataRecord, **kwargs) -> list[str]:
        """
        The list of input fields to be templated into the prompt(s).
        If the user provides a list of "project_cols" in kwargs, then this list will be returned.
        Otherwise, this function returns the list of all field names in the candidate record.

        Args:
            candidate (DataRecord): The input record.
            kwargs: The keyword arguments provided by the user.

        Returns:
            list[str]: The list of input field names.
        """
        return kwargs.get("project_cols", candidate.get_field_names())

    def _get_input_fields_desc(self, candidate: DataRecord, input_fields: list[str]) -> str:
        """
        Returns a multi-line description of each input field for the prompt.

        Args:
            input_fields (list[str]): The input fields.

        Returns:
            str: The input fields description.
        """
        input_fields_desc = ""
        for field_name in input_fields:
            input_fields_desc += f"- {field_name}: {candidate.get_field_type(field_name)._desc}\n"

        return input_fields_desc[:-1]

    def _get_output_fields_desc(self, output_fields: list[str], **kwargs) -> str:
        """
        Returns a multi-line description of each output field for the prompt.

        Args:
            output_fields (list[str]): The output fields.
            kwargs: The keyword arguments provided by the user.

        Returns:
            str: The output fields description.
        """
        output_fields_desc = ""
        output_schema: Schema = kwargs.get("output_schema")
        if self.prompt_strategy.is_convert_prompt():
            assert output_schema is not None, "Output schema must be provided for convert prompts."

            field_desc_map = output_schema.field_desc_map()
            for field_name in output_fields:
                output_fields_desc += f"- {field_name}: {field_desc_map[field_name]}\n"

        # strip the last newline characters from the field descriptions and return
        return output_fields_desc[:-1]

    def _get_filter_condition(self, **kwargs) -> str | None:
        """
        Returns the filter condition for the filter operation.

        Returns:
            str | None: The filter condition (if applicable).
        """
        filter_condition = kwargs.get("filter_condition")
        if self.prompt_strategy.is_bool_prompt():
            assert filter_condition is not None, "Filter condition must be provided for filter operations."

        return filter_condition

    def _get_original_output(self, **kwargs) -> str | None:
        """
        Returns the original output from a previous model generation for the critique and refinement operations.

        Args:
            kwargs: The keyword arguments provided by the user.

        Returns:
            str | None: The original output.
        """
        original_output = kwargs.get("original_output")
        if self.prompt_strategy.is_critic_prompt() or self.prompt_strategy.is_refine_prompt():
            assert original_output is not None, (
                "Original output must be provided for critique and refinement operations."
            )

        return original_output

    def _get_critique_output(self, **kwargs) -> str | None:
        """
        Returns the critique output for the refinement operation.

        Args:
            kwargs: The keyword arguments provided by the user.

        Returns:
            str | None: The critique output.
        """
        critique_output = kwargs.get("critique_output")
        if self.prompt_strategy.is_refine_prompt():
            assert critique_output is not None, "Critique output must be provided for refinement operations."

        return critique_output

    def _get_model_responses(self, **kwargs) -> str | None:
        """
        Returns the model responses for the mixture-of-agents aggregation operation.

        Args:
            kwargs: The keyword arguments provided by the user.

        Returns:
            str | None: The model responses.
        """
        model_responses = None
        if self.prompt_strategy.is_moa_aggregator_prompt():
            model_responses = ""
            for idx, model_response in enumerate(kwargs.get("model_responses")):
                model_responses += f"MODEL RESPONSE {idx + 1}: {model_response}\n"

        return model_responses

    def _get_chunk_outputs(self, **kwargs) -> str | None:
        """
        Returns the chunk outputs for the split-convert.

        Args:
            kwargs: The keyword arguments provided by the user.

        Returns:
            str | None: The chunk outputs.
        """
        chunk_outputs = None
        if self.prompt_strategy.is_split_merger_prompt():
            chunk_outputs = ""
            for idx, chunk_output in enumerate(kwargs.get("chunk_outputs")):
                chunk_outputs += f"CHUNK OUTPUT {idx + 1}: {chunk_output}\n"

        return chunk_outputs

    def _get_output_format_instruction(self) -> str:
        """
        Returns the output format instruction based on the cardinality.

        Returns:
            str: The output format instruction.
        """
        return (
            ONE_TO_ONE_OUTPUT_FORMAT_INSTRUCTION
            if self.cardinality == Cardinality.ONE_TO_ONE
            else ONE_TO_MANY_OUTPUT_FORMAT_INSTRUCTION
        )

    def _get_job_instruction(self) -> str | None:
        """
        Returns the job instruction based on the prompt strategy.

        Returns:
            str | None: The job instruction (if applicable).
        """
        prompt_strategy_to_job_instruction = {
            PromptStrategy.COT_BOOL: COT_BOOL_JOB_INSTRUCTION,
            PromptStrategy.COT_BOOL_IMAGE: COT_BOOL_IMAGE_JOB_INSTRUCTION,
            PromptStrategy.COT_QA: COT_QA_JOB_INSTRUCTION,
            PromptStrategy.COT_QA_IMAGE: COT_QA_IMAGE_JOB_INSTRUCTION,
            PromptStrategy.COT_MOA_PROPOSER: COT_MOA_PROPOSER_JOB_INSTRUCTION,
            PromptStrategy.COT_MOA_PROPOSER_IMAGE: COT_MOA_PROPOSER_IMAGE_JOB_INSTRUCTION,
            PromptStrategy.SPLIT_PROPOSER: SPLIT_PROPOSER_JOB_INSTRUCTION,
        }
        return prompt_strategy_to_job_instruction.get(self.prompt_strategy)

    def _get_critique_criteria(self) -> str | None:
        """
        Returns the critique criteria for the critique operation.

        Returns:
            str | None: The critique criteria (if applicable).
        """
        critique_criteria = None
        if self.prompt_strategy.is_critic_prompt():
            critique_criteria = (
                COT_QA_IMAGE_CRITIQUE_CRITERIA if self.prompt_strategy.is_image_prompt() else COT_QA_CRITIQUE_CRITERIA
            )

        return critique_criteria

    def _get_refinement_criteria(self) -> str | None:
        """
        Returns the refinement criteria for the refinement operation.

        Returns:
            str | None: The refinement criteria (if applicable).
        """
        refinement_criteria = None
        if self.prompt_strategy.is_refine_prompt():
            refinement_criteria = (
                COT_QA_IMAGE_REFINEMENT_CRITERIA
                if self.prompt_strategy.is_image_prompt()
                else COT_QA_REFINEMENT_CRITERIA
            )

        return refinement_criteria

    def _get_finish_instruction(self) -> str | None:
        """
        Returns the finish instruction for the critique and refinement operations.

        Returns:
            str | None: The finish instruction (if applicable).
        """
        finish_instruction = None
        if self.prompt_strategy.is_critic_prompt():
            finish_instruction = COT_QA_CRITIQUE_FINISH_INSTRUCTION
        elif self.prompt_strategy.is_refine_prompt():
            finish_instruction = COT_QA_REFINEMENT_FINISH_INSTRUCTION

        return finish_instruction

    def _get_example_input_fields(self) -> str | None:
        """
        Returns the example input fields for the prompt.

        Returns:
            str | None: The example input fields (if applicable).
        """
        prompt_strategy_to_example_input_fields = {
            PromptStrategy.COT_BOOL: COT_BOOL_EXAMPLE_INPUT_FIELDS,
            PromptStrategy.COT_BOOL_IMAGE: COT_BOOL_IMAGE_EXAMPLE_INPUT_FIELDS,
            PromptStrategy.COT_QA: COT_QA_EXAMPLE_INPUT_FIELDS,
            PromptStrategy.COT_QA_IMAGE: COT_QA_IMAGE_EXAMPLE_INPUT_FIELDS,
            PromptStrategy.COT_MOA_PROPOSER: COT_MOA_PROPOSER_EXAMPLE_INPUT_FIELDS,
            PromptStrategy.COT_MOA_PROPOSER_IMAGE: COT_MOA_PROPOSER_IMAGE_EXAMPLE_INPUT_FIELDS,
            PromptStrategy.SPLIT_PROPOSER: SPLIT_PROPOSER_EXAMPLE_INPUT_FIELDS,
        }

        return prompt_strategy_to_example_input_fields.get(self.prompt_strategy)

    def _get_example_output_fields(self) -> str | None:
        """
        Returns the example output fields for the prompt.

        Returns:
            str | None: The example output fields (if applicable).
        """
        prompt_strategy_to_example_output_fields = {
            PromptStrategy.COT_QA: COT_QA_EXAMPLE_OUTPUT_FIELDS,
            PromptStrategy.COT_QA_IMAGE: COT_QA_IMAGE_EXAMPLE_OUTPUT_FIELDS,
            PromptStrategy.COT_MOA_PROPOSER: COT_MOA_PROPOSER_EXAMPLE_OUTPUT_FIELDS,
            PromptStrategy.COT_MOA_PROPOSER_IMAGE: COT_MOA_PROPOSER_IMAGE_EXAMPLE_OUTPUT_FIELDS,
            PromptStrategy.SPLIT_PROPOSER: SPLIT_PROPOSER_EXAMPLE_OUTPUT_FIELDS,
        }

        return prompt_strategy_to_example_output_fields.get(self.prompt_strategy)

    def _get_example_context(self) -> str | None:
        """
        Returns the example context for the prompt.

        Returns:
            str | None: The example context (if applicable).
        """
        prompt_strategy_to_example_context = {
            PromptStrategy.COT_BOOL: COT_BOOL_EXAMPLE_CONTEXT,
            PromptStrategy.COT_BOOL_IMAGE: COT_BOOL_IMAGE_EXAMPLE_CONTEXT,
            PromptStrategy.COT_QA: COT_QA_EXAMPLE_CONTEXT,
            PromptStrategy.COT_QA_IMAGE: COT_QA_IMAGE_EXAMPLE_CONTEXT,
            PromptStrategy.COT_MOA_PROPOSER: COT_MOA_PROPOSER_EXAMPLE_CONTEXT,
            PromptStrategy.COT_MOA_PROPOSER_IMAGE: COT_MOA_PROPOSER_IMAGE_EXAMPLE_CONTEXT,
            PromptStrategy.SPLIT_PROPOSER: SPLIT_PROPOSER_EXAMPLE_CONTEXT,
        }

        return prompt_strategy_to_example_context.get(self.prompt_strategy)

    def _get_image_disclaimer(self) -> str:
        """
        Returns the image disclaimer for the prompt. The disclaimer must be an empty string
        for text prompts.

        Returns:
            str: The image disclaimer. If this is a text prompt then it is an empty string.
        """
        prompt_strategy_to_image_disclaimer = {
            PromptStrategy.COT_BOOL_IMAGE: COT_BOOL_IMAGE_DISCLAIMER,
            PromptStrategy.COT_QA_IMAGE: COT_QA_IMAGE_DISCLAIMER,
            PromptStrategy.COT_MOA_PROPOSER_IMAGE: COT_MOA_PROPOSER_IMAGE_DISCLAIMER,
        }

        return prompt_strategy_to_image_disclaimer.get(self.prompt_strategy, "")

    def _get_example_filter_condition(self) -> str | None:
        """
        Returns the example filter condition for the prompt.

        Returns:
            str | None: The example filter condition (if applicable).
        """
        prompt_strategy_to_example_filter_condition = {
            PromptStrategy.COT_BOOL: COT_BOOL_EXAMPLE_FILTER_CONDITION,
            PromptStrategy.COT_BOOL_IMAGE: COT_BOOL_IMAGE_EXAMPLE_FILTER_CONDITION,
        }

        return prompt_strategy_to_example_filter_condition.get(self.prompt_strategy)

    def _get_example_reasoning(self) -> str | None:
        """
        Returns the example reasoning for the prompt.

        Returns:
            str | None: The example reasoning (if applicable).
        """
        prompt_strategy_to_example_reasoning = {
            PromptStrategy.COT_BOOL: COT_BOOL_EXAMPLE_REASONING,
            PromptStrategy.COT_BOOL_IMAGE: COT_BOOL_IMAGE_EXAMPLE_REASONING,
            PromptStrategy.COT_QA: COT_QA_EXAMPLE_REASONING,
            PromptStrategy.COT_QA_IMAGE: COT_QA_IMAGE_EXAMPLE_REASONING,
        }

        return prompt_strategy_to_example_reasoning.get(self.prompt_strategy)

    def _get_example_answer(self) -> str | None:
        """
        Returns the example answer for the prompt.

        Returns:
            str | None: The example answer (if applicable).
        """
        prompt_strategy_to_example_answer = {
            PromptStrategy.COT_QA: COT_QA_EXAMPLE_ANSWER,
            PromptStrategy.COT_QA_IMAGE: COT_QA_IMAGE_EXAMPLE_ANSWER,
            PromptStrategy.COT_MOA_PROPOSER: COT_MOA_PROPOSER_EXAMPLE_ANSWER,
            PromptStrategy.COT_MOA_PROPOSER_IMAGE: COT_MOA_PROPOSER_IMAGE_EXAMPLE_ANSWER,
            PromptStrategy.SPLIT_PROPOSER: SPLIT_PROPOSER_EXAMPLE_ANSWER,
        }

        return prompt_strategy_to_example_answer.get(self.prompt_strategy)

    def _get_all_format_kwargs(
        self, candidate: DataRecord, input_fields: list[str], output_fields: list[str], **kwargs
    ) -> dict:
        """
        Returns a dictionary containing all the format kwargs for templating the prompts.

        Args:
            candidate (DataRecord): The input record.
            input_fields (list[str]): The input fields.
            output_fields (list[str]): The output fields.
            kwargs: The keyword arguments provided by the user.

        Returns:
            dict: The dictionary containing all the format kwargs.
        """
        # get format kwargs which depend on the input data
        input_format_kwargs = {
            "context": self._get_context(candidate, input_fields),
            "input_fields_desc": self._get_input_fields_desc(candidate, input_fields),
            "output_fields_desc": self._get_output_fields_desc(output_fields, **kwargs),
            "filter_condition": self._get_filter_condition(**kwargs),
            "original_output": self._get_original_output(**kwargs),
            "critique_output": self._get_critique_output(**kwargs),
            "model_responses": self._get_model_responses(**kwargs),
            "chunk_outputs": self._get_chunk_outputs(**kwargs),
        }

        # get format kwargs which depend on the prompt strategy
        prompt_strategy_format_kwargs = {
            "output_format_instruction": self._get_output_format_instruction(),
            "job_instruction": self._get_job_instruction(),
            "critique_criteria": self._get_critique_criteria(),
            "refinement_criteria": self._get_refinement_criteria(),
            "finish_instruction": self._get_finish_instruction(),
            "example_input_fields": self._get_example_input_fields(),
            "example_output_fields": self._get_example_output_fields(),
            "example_context": self._get_example_context(),
            "image_disclaimer": self._get_image_disclaimer(),
            "example_filter_condition": self._get_example_filter_condition(),
            "example_reasoning": self._get_example_reasoning(),
            "example_answer": self._get_example_answer(),
        }

        # return all format kwargs
        return {**input_format_kwargs, **prompt_strategy_format_kwargs}

    def _create_image_messages(self, candidate: DataRecord, input_fields: list[str]) -> list[dict]:
        """
        Parses the candidate record and returns the image messages for the chat payload.

        Args:
            candidate (DataRecord): The input record.
            input_fields (list[str]): The list of input fields.

        Returns:
            list[dict]: The image messages for the chat payload.
        """
        # create a message for each image in an input field with an image (or list of image) type
        image_messages = []
        for field_name in input_fields:
            field_value = candidate[field_name]
            field_type = candidate.get_field_type(field_name)

            # image filepath (or list of image filepaths)
            if isinstance(field_type, ImageFilepathField):
                with open(field_value, "rb") as f:
                    base64_image_str = base64.b64encode(f.read()).decode("utf-8")
                image_messages.append(
                    {"role": "user", "type": "image", "content": f"data:image/jpeg;base64,{base64_image_str}"}
                )

            elif hasattr(field_type, "element_type") and issubclass(field_type.element_type, ImageFilepathField):
                for image_filepath in field_value:
                    with open(image_filepath, "rb") as f:
                        base64_image_str = base64.b64encode(f.read()).decode("utf-8")
                    image_messages.append(
                        {"role": "user", "type": "image", "content": f"data:image/jpeg;base64,{base64_image_str}"}
                    )

            # image url (or list of image urls)
            elif isinstance(field_type, ImageURLField):
                image_messages.append({"role": "user", "type": "image", "content": field_value})

            elif hasattr(field_type, "element_type") and issubclass(field_type.element_type, ImageURLField):
                for image_url in field_value:
                    image_messages.append({"role": "user", "type": "image", "content": image_url})

            # pre-encoded images (or list of pre-encoded images)
            elif isinstance(field_type, ImageBase64Field):
                base64_image_str = field_value.decode("utf-8")
                image_messages.append(
                    {"role": "user", "type": "image", "content": f"data:image/jpeg;base64,{base64_image_str}"}
                )

            elif hasattr(field_type, "element_type") and issubclass(field_type.element_type, ImageBase64Field):
                for base64_image in field_value:
                    base64_image_str = base64_image.decode("utf-8")
                    image_messages.append(
                        {"role": "user", "type": "image", "content": f"data:image/jpeg;base64,{base64_image_str}"}
                    )

        return image_messages

    def _get_system_prompt(self, **format_kwargs) -> str | None:
        """
        Returns the fully templated system prompt for the given prompt strategy.
        Returns None if the prompt strategy does not use a system prompt.

        Returns:
            str | None: The fully templated system prompt (or None if the prompt strategy
                does not use a system prompt).
        """
        base_prompt: str = self.BASE_SYSTEM_PROMPT_MAP.get(self.prompt_strategy)

        # for critic and refine prompt strategies, we do not use a base prompt
        if base_prompt is None:
            return base_prompt

        return base_prompt.format(**format_kwargs)

    def _get_user_messages(self, candidate: DataRecord, input_fields: list[str], **kwargs) -> str:
        """
        Returns a list of messages for the chat payload based on the prompt strategy.

        Args:
            candidate (DataRecord): The input record.
            input_fields (list[str]): The input fields.
            output_fields (list[str]): The output fields.
            kwargs: The formatting kwargs and some keyword arguments provided by the user.

        Returns:
            Tuple[str, str | None]: The fully templated start and end of the user prompt.
                The second element will be None for text prompts.
        """
        # get the base prompt template
        base_prompt = self.BASE_USER_PROMPT_MAP.get(self.prompt_strategy)

        # get any image messages for the chat payload (will be an empty list if this is not an image prompt)
        image_messages = (
            self._create_image_messages(candidate, input_fields) if self.prompt_strategy.is_image_prompt() else []
        )

        # get any original messages for critique and refinement operations
        original_messages = kwargs.get("original_messages")
        if self.prompt_strategy.is_critic_prompt() or self.prompt_strategy.is_refine_prompt():
            assert original_messages is not None, (
                "Original messages must be provided for critique and refinement operations."
            )

        # construct the user messages based on the prompt strategy
        user_messages = []
        if self.prompt_strategy.is_critic_prompt() or self.prompt_strategy.is_refine_prompt():
            # NOTE: if this critic / refinement prompt is processing images, those images will
            #       be part of the `original_messages` and will show up in the final chat payload
            base_prompt_start, base_prompt_end = base_prompt.split("<<original-prompt-placeholder>>\n")
            user_messages.append({"role": "user", "type": "text", "content": base_prompt_start.format(**kwargs)})
            user_messages.extend(original_messages)
            user_messages.append({"role": "user", "type": "text", "content": base_prompt_end.format(**kwargs)})

        elif self.prompt_strategy.is_image_prompt():
            base_prompt_start, base_prompt_end = base_prompt.split("<<image-placeholder>>\n")
            user_messages.append({"role": "user", "type": "text", "content": base_prompt_start.format(**kwargs)})
            user_messages.extend(image_messages)
            user_messages.append({"role": "user", "type": "text", "content": base_prompt_end.format(**kwargs)})

        else:
            base_prompt = base_prompt.replace("<<image-placeholder>>", "")
            user_messages.append({"role": "user", "type": "text", "content": base_prompt.format(**kwargs)})

        return user_messages

    def _process_custom_user_prompt(self, candidate: DataRecord, input_fields: list[str], **kwargs) -> list[dict]:
        """
        Processes a custom user prompt provided by the user.

        Args:
            candidate (DataRecord): The input record.
            kwargs: The keyword arguments provided by the user.

        Returns:
            list[dict]: The messages for the chat payload.
        """
        # get the user prompt
        user_prompt: str = kwargs["prompt"]

        # sanity check that we have all the inputs for the user's prompt template
        prompt_field_names = [fname for _, fname, _, _ in Formatter().parse(user_prompt) if fname]
        fields_check = all([field in input_fields for field in prompt_field_names])
        if not fields_check:
            if sorted(candidate.get_field_names()) != (input_fields):
                err_msg = (
                    f"Prompt string has fields which are not in input fields.\n"
                    f"Prompt fields: {prompt_field_names}\n"
                    f"Computed fields: {candidate.get_field_names()}\n"
                    f"Input fields: {input_fields}\n"
                    f"Be careful that you are not projecting out computed fields. "
                    f"If you use `depends_on` in your program, make sure it includes the fields you need."
                )
            else:
                err_msg = (
                    f"Prompt string has fields which are not in input fields.\n"
                    f"Prompt fields: {prompt_field_names}\n"
                    f"Input fields: {input_fields}\n"
                )
            assert fields_check, err_msg

        # build set of format kwargs
        format_kwargs = {
            field_name: "<bytes>"
            if isinstance(candidate.get_field_type(field_name), BytesField)
            else candidate[field_name]
            for field_name in input_fields
        }

        # split prompt on <<image-placeholder>> if it exists
        if "<<image-placeholder>>" in user_prompt:
            raise NotImplementedError("Image prompts are not yet supported.")

        prompt_sections = user_prompt.split("<<image-placeholder>>")
        messages = [{"role": "user", "type": "text", "content": prompt_sections[0].format(**format_kwargs)}]

        # NOTE: this currently assumes that the user can only provide a single <<image-placeholder>>
        if len(prompt_sections) > 1:
            image_messages = self._create_image_messages(candidate, input_fields)
            messages.extend(image_messages)
            messages.append({"role": "user", "type": "text", "content": prompt_sections[1].format(**format_kwargs)})

        return messages

    def create_messages(self, candidate: DataRecord, output_fields: list[str], **kwargs) -> list[dict]:
        """
        Creates the messages for the chat payload based on the prompt strategy.

        Each message will be a dictionary with the following format:
        {
            "role": "user" | "system",
            "type": "text" | "image",
            "content": str
        }

        Args:
            candidate (DataRecord): The input record.
            output_fields (list[str]): The output fields.
            kwargs: The keyword arguments provided by the user.

        Returns:
            list[dict]: The messages for the chat payload.
        """
        # compute the set of input fields
        input_fields = self._get_input_fields(candidate, **kwargs)

        # if the user provides a prompt, we process that prompt into messages and return them
        if "prompt" in kwargs:
            messages = []
            if "system_prompt" in kwargs:
                messages.append({"role": "system", "type": "text", "content": kwargs["system_prompt"]})
            messages.extend(self._process_custom_user_prompt(candidate, input_fields, **kwargs))
            return messages

        # initialize messages
        messages = []

        # compute the full dictionary of format kwargs and add to kwargs
        format_kwargs = self._get_all_format_kwargs(candidate, input_fields, output_fields, **kwargs)
        kwargs = {**kwargs, **format_kwargs}

        # generate system message (if applicable)
        system_prompt = self._get_system_prompt(**kwargs)
        if system_prompt is not None:
            messages.append({"role": "system", "type": "text", "content": system_prompt})

        # generate user messages and add to messages
        user_messages = self._get_user_messages(candidate, input_fields, **kwargs)
        messages.extend(user_messages)

        return messages



from __future__ import annotations

import json
from collections.abc import Generator
from typing import Any

import pandas as pd

from palimpzest.core.data.dataclasses import ExecutionStats, PlanStats, RecordOpStats
from palimpzest.core.lib.fields import Field
from palimpzest.core.lib.schemas import Schema
from palimpzest.utils.hash_helpers import hash_for_id


class DataRecord:
    """A DataRecord is a single record of data matching some Schema."""

    def __init__(
        self,
        schema: Schema,
        source_idx: int,
        parent_id: str | None = None,
        cardinality_idx: int | None = None,
    ):
        # check that source_idx is provided
        assert source_idx is not None, "Every DataRecord must be constructed with a source_idx"

        # schema for the data record
        self.schema = schema

        # mapping from field names to Field objects; effectively a mapping from a field name to its type        
        self.field_types: dict[str, Field] = schema.field_map()

        # mapping from field names to their values
        self.field_values: dict[str, Any] = {}

        # the index in the DataReader from which this DataRecord is derived
        self.source_idx = int(source_idx)

        # the id of the parent record(s) from which this DataRecord is derived
        self.parent_id = parent_id

        # store the cardinality index
        self.cardinality_idx = cardinality_idx

        # indicator variable which may be flipped by filter operations to signal when a record has been filtered out
        self.passed_operator = True

        # NOTE: Record ids are hashed based on:
        # 0. their schema (keys)
        # 1. their parent record id(s) (or source_idx if there is no parent record)
        # 2. their index in the fan out (if this is in a one-to-many operation)
        #
        # We currently do NOT hash just based on record content (i.e. schema (key, value) pairs)
        # because multiple outputs for a given operation may have the exact same
        # schema (key, value) pairs.
        #
        # We may revisit this hashing scheme in the future.

        # unique identifier for the record
        id_str = (
            str(schema) + (parent_id if parent_id is not None else str(self.source_idx))
            if cardinality_idx is None
            else str(schema) + str(cardinality_idx) + str(parent_id if parent_id is not None else str(self.source_idx))
        )
        # TODO(Jun): build-in id should has a special name, the current self.id is too general which would conflict with user defined schema too easily.
        # the options: built_in_id, generated_id
        self.id = hash_for_id(id_str)


    def __setattr__(self, name: str, value: Any, /) -> None:
        if name in ["schema", "field_types", "field_values", "source_idx", "parent_id", "cardinality_idx", "passed_operator", "id"]:
            super().__setattr__(name, value)
        else:
            self.field_values[name] = value


    def __getattr__(self, name: str) -> Any:
        if name == "field_values":
            pass
        elif name in self.field_values:
            return self.field_values[name]
        else:
            raise AttributeError(f"'{type(self).__name__}' object has no attribute '{name}'")


    def __getitem__(self, field: str) -> Any:
        return self.__getattr__(field)


    def __setitem__(self, field: str, value: Any) -> None:
        self.__setattr__(field, value)


    def __str__(self, truncate: int | None = 15) -> str:
        if truncate is not None:
            items = (f"{k}={str(v)[:truncate]!r}{'...' if len(str(v)) > truncate else ''}" for k, v in sorted(self.field_values.items()))
        else:
            items = (f"{k}={v!r}" for k, v in sorted(self.field_values.items()))
        return "{}({})".format(type(self).__name__, ", ".join(items))

    def __repr__(self) -> str:
        return self.__str__(truncate=None)

    def __eq__(self, other):
        return isinstance(other, DataRecord) and self.field_values == other.field_values and self.schema.get_desc() == other.schema.get_desc()


    def __hash__(self):
        return hash(self.to_json_str(bytes_to_str=True))


    def __iter__(self):
        yield from self.field_values.items()


    def get_field_names(self):
        return list(self.field_values.keys())


    def get_field_type(self, field_name: str) -> Field:
        return self.field_types[field_name]


    def copy(self, include_bytes: bool = True, project_cols: list[str] | None = None):
        # make new record which has parent_record as its parent (and the same source_idx)
        new_dr = DataRecord(
            self.schema,
            source_idx=self.source_idx,
            parent_id=self.id,
            cardinality_idx=self.cardinality_idx,
        )

        # copy the passed_operator attribute
        new_dr.passed_operator = self.passed_operator

        # get the set of fields to copy from the parent record
        copy_field_names = project_cols if project_cols is not None else self.get_field_names()
        copy_field_names = [field.split(".")[-1] for field in copy_field_names]

        # copy field types and values from the parent
        for field_name in copy_field_names:
            field_type = self.get_field_type(field_name)
            field_value = self[field_name]
            if (
                not include_bytes
                and isinstance(field_value, bytes)
                or (isinstance(field_value, list) and len(field_value) > 0 and isinstance(field_value[0], bytes))
            ):
                continue

            # set field and value
            new_dr.field_types[field_name] = field_type
            new_dr[field_name] = field_value

        return new_dr


    @staticmethod
    def from_parent(
        schema: Schema,
        parent_record: DataRecord,
        project_cols: list[str] | None = None,
        cardinality_idx: int | None = None,
    ) -> DataRecord:
        # project_cols must be None or contain at least one column
        assert project_cols is None or len(project_cols) > 1, "must have at least one column if using projection"

        # if project_cols is None, then the new schema is a union of the provided schema and parent_record.schema;
        # otherwise, it's a ProjectSchema
        new_schema = schema.union(parent_record.schema)
        if project_cols is not None:
            new_schema = new_schema.project(project_cols)

        # make new record which has parent_record as its parent (and the same source_idx)
        new_dr = DataRecord(
            new_schema,
            source_idx=parent_record.source_idx,
            parent_id=parent_record.id,
            cardinality_idx=cardinality_idx,
        )

        # get the set of fields and field descriptions to copy from the parent record
        copy_field_names = project_cols if project_cols is not None else parent_record.get_field_names()
        copy_field_names = [field.split(".")[-1] for field in copy_field_names]

        # copy fields from the parent
        for field_name in copy_field_names:
            new_dr.field_types[field_name] = parent_record.get_field_type(field_name)
            new_dr[field_name] = parent_record[field_name]

        return new_dr


    @staticmethod
    def from_agg_parents(
        schema: Schema,
        parent_records: DataRecordSet,
        project_cols: list[str] | None = None,
        cardinality_idx: int | None = None,
    ) -> DataRecord:
        # TODO: we can implement this once we support having multiple parent ids
        pass


    @staticmethod
    def from_join_parents(
        left_schema: Schema,
        right_schema: Schema,
        left_parent_record: DataRecord,
        right_parent_record: DataRecord,
        project_cols: list[str] | None = None,
        cardinality_idx: int = None,
    ) -> DataRecord:
        # TODO: we can implement this method if/when we add joins
        pass


    @staticmethod
    def from_df(df: pd.DataFrame, schema: Schema | None = None) -> list[DataRecord]:
        """Create a list of DataRecords from a pandas DataFrame
        
        Args:
            df (pd.DataFrame): Input DataFrame
            schema (Schema, optional): Schema for the DataRecords. If None, will be derived from DataFrame  
        
        Returns:
            list[DataRecord]: List of DataRecord instances
        """
        if df is None:
            raise ValueError("DataFrame is None!")

        records = []
        if schema is None:
            schema = Schema.from_df(df)

        field_map = schema.field_map()
        for source_idx, row in df.iterrows():
            row_dict = row.to_dict()
            record = DataRecord(schema=schema, source_idx=source_idx)
            record.field_values = row_dict
            record.field_types = {field_name: field_map[field_name] for field_name in row_dict}
            records.append(record)

        return records

    @staticmethod
    def to_df(records: list[DataRecord], project_cols: list[str] | None = None) -> pd.DataFrame:
        if len(records) == 0:
            return pd.DataFrame()

        fields = records[0].get_field_names()
        if project_cols is not None and len(project_cols) > 0:
            fields = [field for field in fields if field in project_cols]

        return pd.DataFrame([
            {k: record[k] for k in fields}
            for record in records
        ])

    def to_json_str(self, include_bytes: bool = True, bytes_to_str: bool = False, project_cols: list[str] | None = None):
        """Return a JSON representation of this DataRecord"""
        record_dict = self.to_dict(include_bytes, bytes_to_str, project_cols)
        record_dict = {
            field_name: self.schema.field_to_json(field_name, field_value)
            for field_name, field_value in record_dict.items()
        }
        return json.dumps(record_dict, indent=2)

    def to_dict(self, include_bytes: bool = True, bytes_to_str: bool = False, project_cols: list[str] | None = None):
        """Return a dictionary representation of this DataRecord"""
        # TODO(chjun): In case of numpy types, the json.dumps will fail. Convert to native types.
        # Better ways to handle this.
        dct = pd.Series(self.field_values).to_dict()

        if project_cols is not None and len(project_cols) > 0:
            project_field_names = set(field.split(".")[-1] for field in project_cols)
            dct = {k: v for k, v in dct.items() if k in project_field_names}

        if not include_bytes:
            for k, v in dct.items():
                if isinstance(v, bytes) or (isinstance(v, list) and len(v) > 0 and  any([isinstance(elt, bytes) for elt in v])):
                    dct[k] = "<bytes>"

        if bytes_to_str:
            for k, v in dct.items():
                if isinstance(v, bytes):
                    dct[k] = v.decode("utf-8")
                elif isinstance(v, list) and len(v) > 0 and any([isinstance(elt, bytes) for elt in v]):
                    dct[k] = [elt.decode("utf-8") if isinstance(elt, bytes) else elt for elt in v]

        return dct


class DataRecordSet:
    """
    A DataRecordSet contains a list of DataRecords that share the same schema, same parent_id, and same source_idx.

    We explicitly check that this is True.

    The record_op_stats could be empty if the DataRecordSet is not from executing an operator.
    """
    def __init__(
            self,
            data_records: list[DataRecord],
            record_op_stats: list[RecordOpStats],
            field_to_score_fn: dict[str, str | callable] | None = None,
        ):
        # check that all data_records are derived from the same parent record
        if len(data_records) > 0:
            parent_id = data_records[0].parent_id
            error_msg = "DataRecordSet must be constructed from the output of executing a single operator on a single input."
            assert all([dr.parent_id == parent_id for dr in data_records]), error_msg

        # set data_records, parent_id, and source_idx; note that it is possible for
        # data_records to be an empty list in the event of a failed convert
        self.data_records = data_records
        self.parent_id = data_records[0].parent_id if len(data_records) > 0 else None
        self.source_idx = data_records[0].source_idx if len(data_records) > 0 else None
        self.schema = data_records[0].schema if len(data_records) > 0 else None

        # set statistics for generating these records
        self.record_op_stats = record_op_stats

        # assign field_to_score_fn if provided
        self.field_to_score_fn = {} if field_to_score_fn is None else field_to_score_fn

    def get_total_cost(self) -> float:
        return sum([record_op_stats.cost_per_record for record_op_stats in self.record_op_stats])

    def get_field_to_score_fn(self) -> dict[str, str | callable]:
        return self.field_to_score_fn

    def __getitem__(self, slice) -> DataRecord | list[DataRecord]:
        return self.data_records[slice]

    def __len__(self) -> int:
        return len(self.data_records)

    def __iter__(self) -> Generator[DataRecord]:
        yield from self.data_records


class DataRecordCollection:
    """
    A DataRecordCollection contains a list of DataRecords.

    This is a wrapper class for list[DataRecord] to support more advanced features for output of execute().

    The difference between DataRecordSet and DataRecordCollection 

    Goal: 
        DataRecordSet is a set of DataRecords that share the same schema, same parent_id, and same source_idx.
        DataRecordCollection is a general wrapper for list[DataRecord].
    
    Usage:
        DataRecordSet is used for the output of executing an operator.
        DataRecordCollection is used for the output of executing a query, we definitely could extend it to support more advanced features for output of execute().
    """
    # TODO(Jun): consider to have stats_manager class to centralize stats management.
    def __init__(self, data_records: list[DataRecord], execution_stats: ExecutionStats | None = None, plan_stats: PlanStats | None = None):
        self.data_records = data_records
        self.execution_stats = execution_stats
        self.plan_stats = plan_stats
        self.executed_plans = self._get_executed_plans()

    def __iter__(self) -> Generator[DataRecord]:
        """Allow iterating directly over the data records"""
        yield from self.data_records

    def __len__(self):
        """Return the number of records in the collection"""
        return len(self.data_records)

    def to_df(self, cols: list[str] | None = None):
        return DataRecord.to_df(self.data_records, cols)

    def _get_executed_plans(self):
        if self.plan_stats is not None:
            return [self.plan_stats.plan_str]
        elif self.execution_stats is not None:
            return list(self.execution_stats.plan_strs.values())
        else:
            return None



import types

from palimpzest.core.lib.fields import (
    BooleanField,
    BytesField,
    Field,
    FloatField,
    IntField,
    ListField,
    NumericField,
    StringField,
)


def assert_valid_field_type(field_type: type | types.UnionType | types.GenericAlias | Field) -> str:
    """
    Assert that the field is a valid field type. Return "pz_type" if field_type is a PZ type
    and "python_type" if it is a Python type.
    """
    try:
        assert issubclass(field_type, Field), "type must be a Python type or palimpzest.core.lib.fields.Field"
        return "pz_type"
    except Exception:
        assert isinstance(field_type, (type, types.UnionType, types.GenericAlias)), "type must be a Python type or palimpzest.core.lib.fields.Field"
    
    return "python_type"


def construct_field_type(field_type: type | types.UnionType | types.GenericAlias | Field, desc: str) -> Field:
    """Convert a field type and description to the corresponding PZ field.

    Args:
        type: type for the field (e.g. str, bool, list[int], StringField, etc.)
        desc: description used in the field constructor

    Returns:
        Corresponding Field class

    Raises:
        ValueError: If the type is not recognized
    """
    # if field_type is a PZ type, construct and return the field
    if assert_valid_field_type(field_type) == "pz_type":
        return field_type(desc=desc)

    # otherwise, map the Python type to a PZ type and construct the field
    supported_types_map = {
        str: StringField,
        bool: BooleanField,
        int: IntField,
        float: FloatField,
        int | float: NumericField,
        bytes: BytesField,
        list[str]: ListField(StringField),
        list[bool]: ListField(BooleanField),
        list[int]: ListField(IntField),
        list[float]: ListField(FloatField),
        list[int | float]: ListField(NumericField),
        list[bytes]: ListField(BytesField),
    }

    if field_type not in supported_types_map:
        raise ValueError(f"Unsupported type: {field_type}. Supported types are: {list(supported_types_map.keys())}")

    # get the field class and (if applicable) element field class
    field_cls = supported_types_map[field_type]

    # construct and return the field
    return field_cls(desc=desc)



"""This file contains prompts for convert operations."""

### BASE PROMPTS ###
COT_QA_BASE_SYSTEM_PROMPT = """You are a helpful assistant whose job is to {job_instruction}.
You will be presented with a context and a set of output fields to generate. Your task is to generate a JSON object which fills in the output fields with the correct values.
You will be provided with a description of each input field and each output field. All of the fields in the output JSON object can be derived using information from the context.

{output_format_instruction} Finish your response with a newline character followed by ---

An example is shown below:
---
INPUT FIELDS:
{example_input_fields}

OUTPUT FIELDS:
{example_output_fields}

CONTEXT:
{example_context}
{image_disclaimer}
Let's think step-by-step in order to answer the question.

REASONING: {example_reasoning}

ANSWER:
{example_answer}
---
"""

COT_QA_BASE_USER_PROMPT = """You are a helpful assistant whose job is to {job_instruction}.
You will be presented with a context and a set of output fields to generate. Your task is to generate a JSON object which fills in the output fields with the correct values.
You will be provided with a description of each input field and each output field. All of the fields in the output JSON object can be derived using information from the context.

{output_format_instruction} Finish your response with a newline character followed by ---
---
INPUT FIELDS:
{input_fields_desc}

OUTPUT FIELDS:
{output_fields_desc}

CONTEXT:
{context}
<<image-placeholder>>
Let's think step-by-step in order to answer the question.

REASONING: """


### TEMPLATE INPUTS ###
COT_QA_JOB_INSTRUCTION = """generate a JSON object"""
COT_QA_IMAGE_JOB_INSTRUCTION = """analyze input image(s) and/or text in order to produce a JSON object"""

COT_QA_EXAMPLE_INPUT_FIELDS = """- text: a text passage describing a scientist
- birthday: the scientist's birthday"""
COT_QA_IMAGE_EXAMPLE_INPUT_FIELDS = """- image: an image of a scene
- photographer: the photographer of the image"""

COT_QA_EXAMPLE_OUTPUT_FIELDS = """- name: the name of the scientist
- birth_year: the year the scientist was born"""
COT_QA_IMAGE_EXAMPLE_OUTPUT_FIELDS = """- dog_in_image: true if a dog is in the image and false otherwise
- person_in_image: true if a person is in the image and false otherwise"""

COT_QA_EXAMPLE_CONTEXT = """{{
  "text": "Augusta Ada King, Countess of Lovelace, also known as Ada Lovelace, was an English mathematician and writer chiefly known for her work on Charles Babbage's proposed mechanical general-purpose computer, the Analytical Engine. She was the first to recognise that the machine had applications beyond pure calculation.",
  "birthday": "December 10, 1815"
}}"""
COT_QA_IMAGE_EXAMPLE_CONTEXT = """{{
  "image": <bytes>,
  "photographer": "CameraEnthusiast1"
}}"""

COT_QA_IMAGE_DISCLAIMER = """
<image content provided here; assume in this example the image shows a dog and a cat playing>
"""

COT_QA_EXAMPLE_REASONING = """the text passage mentions the scientist's name as "Augusta Ada King, Countess of Lovelace, also known as Ada Lovelace" and the scientist's birthday as "December 10, 1815". Therefore, the name of the scientist is "Augusta Ada King" and the birth year is 1815."""
COT_QA_IMAGE_EXAMPLE_REASONING = """The image shows a dog playing with a cat, so there is a dog in the image. There is no person in the image."""

COT_QA_EXAMPLE_ANSWER = """{{
  "name": "Augusta Ada King",
  "birth_year": 1815
}}"""
COT_QA_IMAGE_EXAMPLE_ANSWER = """{{
  "dog_in_image": true,
  "person_in_image": false
}}"""



from __future__ import annotations

from palimpzest.core.data.dataclasses import OperatorCostEstimates, RecordOpStats
from palimpzest.core.elements.records import DataRecord, DataRecordSet
from palimpzest.query.operators.physical import PhysicalOperator


class ProjectOp(PhysicalOperator):
    def __init__(self, project_cols: list[str], *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.project_cols = sorted(project_cols)

    def __str__(self):
        op = super().__str__()
        op += f"    Project Columns: {self.project_cols}\n"
        return op

    def get_id_params(self):
        id_params = super().get_id_params()
        return {"project_cols": self.project_cols, **id_params}

    def get_op_params(self):
        op_params = super().get_op_params()
        return {"project_cols": self.project_cols, **op_params}

    def naive_cost_estimates(self, source_op_cost_estimates: OperatorCostEstimates) -> OperatorCostEstimates:
        # for now, assume applying the limit takes negligible additional time (and no cost in USD)
        return OperatorCostEstimates(
            cardinality=source_op_cost_estimates.cardinality,
            time_per_record=0,
            cost_per_record=0,
            quality=1.0,
        )

    def __call__(self, candidate: DataRecord) -> DataRecordSet:
        # create new DataRecord with projection applied
        dr = DataRecord.from_parent(schema=candidate.schema, parent_record=candidate, project_cols=self.project_cols)

        # create RecordOpStats object
        record_op_stats = RecordOpStats(
            record_id=dr.id,
            record_parent_id=dr.parent_id,
            record_source_idx=dr.source_idx,
            record_state=dr.to_dict(include_bytes=False),
            full_op_id=self.get_full_op_id(),
            logical_op_id=self.logical_op_id,
            op_name=self.op_name(),
            time_per_record=0.0,
            cost_per_record=0.0,
            op_details={k: str(v) for k, v in self.get_id_params().items()},
        )

        return DataRecordSet([dr], [record_op_stats])



from __future__ import annotations

from typing import Any

from palimpzest.core.lib.fields import Field
from palimpzest.core.lib.schemas import OperatorDerivedSchema, Schema


# signature for a group by aggregate that applies
# group and aggregation to an input tuple
class GroupBySig:
    def __init__(self, group_by_fields: list[str], agg_funcs: list[str], agg_fields: list[str]):
        self.group_by_fields = group_by_fields
        self.agg_funcs = agg_funcs
        self.agg_fields = agg_fields

    def validate_schema(self, input_schema: Schema) -> tuple[bool, str | None]:
        for f in self.group_by_fields:
            if not hasattr(input_schema, f):
                return (False, "Supplied schema has no field " + f)
        for f in self.agg_fields:
            if not hasattr(input_schema, f):
                return (False, "Supplied schema has no field " + f)
        return (True, None)

    def serialize(self) -> dict[str, Any]:
        out = {
            "group_by_fields": self.group_by_fields,
            "agg_funcs": self.agg_funcs,
            "agg_fields": self.agg_fields,
        }
        return out

    def __str__(self) -> str:
        return "GroupBy(" + repr(self.serialize()) + ")"

    def __hash__(self) -> int:
        # custom hash function
        return hash(repr(self.serialize()))

    def __eq__(self, other) -> bool:
        # __eq__ should be defined for consistency with __hash__
        return isinstance(other, GroupBySig) and self.serialize() == other.serialize()

    def get_agg_field_names(self) -> list[str]:
        ops = []
        for i in range(0, len(self.agg_fields)):
            ops.append(self.agg_funcs[i] + "(" + self.agg_fields[i] + ")")
        return ops

    def output_schema(self) -> type[OperatorDerivedSchema]:
        # the output class varies depending on the group by, so here
        # we dynamically construct this output
        schema = type("CustomGroupBy", (OperatorDerivedSchema,), {})

        for g in self.group_by_fields:
            f = Field(desc=g)
            setattr(schema, g, f)
        ops = self.get_agg_field_names()
        for op in ops:
            f = Field(desc=op)
            setattr(schema, op, f)
        return schema



import json
from typing import Any

import regex as re  # Use regex instead of re to used variable length lookbehind

from palimpzest.constants import Cardinality, Model


def get_json_from_answer(answer: str, model: Model, cardinality: Cardinality) -> dict[str, Any]:
    """
    This function parses an LLM response which is supposed to output a JSON object
    and optimistically searches for the substring containing the JSON object.
    """
    # model-specific trimming for LLAMA3 responses
    if model.is_llama_model():
        answer = answer.split("---")[0]
        answer = answer.replace("True", "true")
        answer = answer.replace("False", "false")

    # split off context / excess, which models sometimes output after answer
    answer = answer.split("Context:")[0]
    answer = answer.split("# this is the answer")[0]

    # trim the answer to only include the JSON dictionary
    if cardinality == Cardinality.ONE_TO_ONE:
        if not answer.strip().startswith("{"):
            # Find the start index of the actual JSON string assuming the prefix is followed by the JSON dictionary
            start_index = answer.find("{")
            if start_index != -1:
                # Remove the prefix and any leading characters before the JSON starts
                answer = answer[start_index:]

        if not answer.strip().endswith("}"):
            # Find the end index of the actual JSON string assuming the suffix is preceded by the JSON dictionary
            end_index = answer.rfind("}")
            if end_index != -1:
                # Remove the suffix and any trailing characters after the JSON ends
                answer = answer[: end_index + 1]

    # otherwise, trim the answer to only include the JSON array
    else:
        if not answer.strip().startswith("["):
            # Find the start index of the actual JSON string assuming the prefix is followed by the JSON array
            start_index = answer.find("[")
            if start_index != -1:
                # Remove the prefix and any leading characters before the JSON starts
                answer = answer[start_index:]

        if not answer.strip().endswith("]"):
            # Find the end index of the actual JSON string
            # assuming the suffix is preceded by the JSON object/array
            end_index = answer.rfind("]")
            if end_index != -1:
                # Remove the suffix and any trailing characters after the JSON ends
                answer = answer[: end_index + 1]

    # Handle weird escaped values. I am not sure why the model
    # is returning these, but the JSON parser can't take them
    answer = answer.replace(r"\_", "_")
    answer = answer.replace("\\n", "\n")
    # Remove https and http prefixes to not conflict with comment detection
    # Handle comments in the JSON response. Use regex from // until end of line
    answer = re.sub(r"(?<!https?:)\/\/.*?$", "", answer, flags=re.MULTILINE)
    answer = re.sub(r",\n.*\.\.\.$", "", answer, flags=re.MULTILINE)
    # Sanitize newlines in the JSON response
    answer = answer.replace("\n", " ")

    # finally, parse and return the JSON object; errors are handled by the caller
    return json.loads(answer)



from enum import Enum

from palimpzest.query.optimizer.optimizer_strategy import (
    GreedyStrategy,
    NoOptimizationStrategy,
    ParetoStrategy,
    SentinelStrategy,
)


class OptimizationStrategyType(Enum):
    """
    OptimizationStrategyType determines which (set of) plan(s) the Optimizer
    will return to the Execution layer.
    """
    GREEDY = GreedyStrategy
    PARETO = ParetoStrategy
    SENTINEL = SentinelStrategy
    NONE = NoOptimizationStrategy

    def no_transformation(self) -> bool:
        """
        Return True if this optimization strategy does not transform the logical plan.
        """
        return self in [OptimizationStrategyType.SENTINEL, OptimizationStrategyType.NONE]

    def is_pareto(self) -> bool:
        """
        Return True if this optimization strategy uses Pareto optimization.
        """
        return self == OptimizationStrategyType.PARETO

    def is_not_pareto(self) -> bool:
        """
        Return True if this optimization strategy does not use Pareto optimization.
        """
        return not self.is_pareto()



from threading import Lock

from openai import OpenAI
from together import Together

from palimpzest.constants import APIClient


class APIClientFactory:
    _instances = {}
    _lock = Lock()

    @classmethod
    def get_client(cls, api_client: APIClient, api_key: str):
        """Get a singleton instance of the requested API client."""
        if api_client not in cls._instances:
            with cls._lock:  # Ensure thread safety
                if api_client not in cls._instances:  # Double-check inside the lock
                    cls._instances[api_client] = cls._create_client(api_client, api_key)
        return cls._instances[api_client]

    @staticmethod
    def _create_client(api_client: APIClient, api_key: str):
        """Create a new client instance based on the api_client name."""
        if api_client == APIClient.OPENAI:
            return OpenAI(api_key=api_key)
        elif api_client == APIClient.TOGETHER:
            return Together(api_key=api_key)
        else:
            raise ValueError(f"Unknown api_client: {api_client}")



"""This file contains prompts for Mixture-of-Agents aggregator operations."""

### SYSTEM PROMPTS ###
COT_MOA_AGG_BASE_SYSTEM_PROMPT = """You are a helpful assistant whose job is to generate a JSON object.
You will be presented with one or more outputs produced by a set of models. Your task is to synthesize these responses into a single, high-quality JSON object which fills in the output fields with the correct values.
It is crucial to critically evaluate the information provided in these responses, recognizing that some of it may be biased or incorrect.

You will be provided with a description of each input field and each output field. All of the fields in the output JSON object can be derived using information from the model responses.

{output_format_instruction} Finish your response with a newline character followed by ---

An example is shown below:
---
MODEL RESPONSE 1: the text mentions the scientist's full name "Augusta Ada King, Countess of Lovelace" and states she was an English mathematician who worked on Babbage's Analytical Engine.

MODEL RESPONSE 2: the text passage mentions the scientist's name as "Augusta Ada King, Countess of Lovelace, also known as Ada Lovelace" and the scientist's birthday as "December 10, 1815". Therefore, the name of the scientist is "Augusta Ada King" and the birth year is 1815.

INPUT FIELDS:
- text: a text passage describing a scientist
- birthday: the scientist's birthday

OUTPUT FIELDS:
- name: the name of the scientist
- birth_year: the year the scientist was born

Let's think step-by-step in order to answer the question.

REASONING: Looking at both model responses, they agree that the scientist's formal name is "Augusta Ada King". Model Response 2 correctly extracts the birth year from the birthday field as 1815. The responses are consistent and provide sufficient evidence for these values.

ANSWER:
{{
  "name": "Augusta Ada King",
  "birth_year": 1815
}}
---
"""

### USER / INSTANCE-SPECIFIC PROMPTS ###
COT_MOA_AGG_BASE_USER_PROMPT = """You are a helpful assistant whose job is to generate a JSON object.
You will be presented with one or more outputs produced by a set of models. Your task is to synthesize these responses into a single, high-quality JSON object which fills in the output fields with the correct values.
It is crucial to critically evaluate the information provided in these responses, recognizing that some of it may be biased or incorrect.

You will be provided with a description of each input field and each output field. All of the fields in the output JSON object can be derived using information from the model responses.

{output_format_instruction} Finish your response with a newline character followed by ---
---
{model_responses}

INPUT FIELDS:
{input_fields_desc}

OUTPUT FIELDS:
{output_fields_desc}

Let's think step-by-step in order to answer the question.

REASONING: """



import logging
from abc import abstractmethod

from palimpzest.core.data.dataclasses import PlanStats
from palimpzest.core.data.datareaders import DataReader
from palimpzest.core.elements.records import DataRecord, DataRecordCollection
from palimpzest.policy import Policy
from palimpzest.query.execution.execution_strategy import ExecutionStrategy, SentinelExecutionStrategy
from palimpzest.query.optimizer.optimizer import Optimizer
from palimpzest.sets import Dataset
from palimpzest.utils.hash_helpers import hash_for_id
from palimpzest.utils.model_helpers import get_models

logger = logging.getLogger(__name__)

class QueryProcessor:
    """
    Processes queries through the complete pipeline:
    1. Optimization phase: Plan generation and selection
    2. Execution phase: Plan execution and result collection
    3. Result phase: Statistics gathering and result formatting
    """
    def __init__(
        self,
        dataset: Dataset,
        optimizer: Optimizer,
        execution_strategy: ExecutionStrategy,
        sentinel_execution_strategy: SentinelExecutionStrategy | None,
        num_samples: int | None = None,
        val_datasource: DataReader | None = None,
        scan_start_idx: int = 0,
        cache: bool = False,
        verbose: bool = False,
        progress: bool = True,
        max_workers: int | None = None,
        policy: Policy | None = None,
        available_models: list[str] | None = None,
        **kwargs,
    ):
        """
        Initialize QueryProcessor with optional custom components.
        
        Args:
            dataset: Dataset to process
            TODO
        """
        self.dataset = dataset
        self.optimizer = optimizer
        self.execution_strategy = execution_strategy
        self.sentinel_execution_strategy = sentinel_execution_strategy

        self.num_samples = num_samples
        self.val_datasource = val_datasource
        self.scan_start_idx = scan_start_idx
        self.cache = cache
        self.verbose = verbose
        self.progress = progress
        self.max_workers = max_workers

        self.policy = policy

        self.available_models = available_models
        if self.available_models is None or len(self.available_models) == 0:
            self.available_models = get_models(include_vision=True)

        if self.verbose:
            print("Available models: ", self.available_models)

        logger.info(f"Initialized QueryProcessor {self.__class__.__name__}")
        logger.debug(f"QueryProcessor initialized with config: {self.__dict__}")

    def execution_id(self) -> str:
        """
        Hash of the class parameters.
        """
        id_str = ""
        for attr, value in self.__dict__.items():
            if not attr.startswith("_"):
                id_str += f"{attr}={value},"

        return hash_for_id(id_str)

    def _execute_best_plan(self, dataset: Dataset, optimizer: Optimizer) -> tuple[list[DataRecord], list[PlanStats]]:
        # get the optimal plan according to the optimizer
        plans = optimizer.optimize(dataset)
        final_plan = plans[0]

        # execute the plan
        records, plan_stats = self.execution_strategy.execute_plan(plan=final_plan)

        # return the output records and plan stats
        return records, [plan_stats]

    # TODO: consider to support dry_run.
    @abstractmethod
    def execute(self) -> DataRecordCollection:
        raise NotImplementedError("Abstract method to be overwritten by sub-classes")



"""This file contains prompts for SplitConvert aggregator operations."""

### SYSTEM PROMPTS ###
COT_SPLIT_MERGER_BASE_SYSTEM_PROMPT = """You are a helpful assistant whose job is to generate a JSON object.
You will be presented with one or more outputs produced by a set of models operating on chunks of an input. Your task is to synthesize these responses into a single, high-quality JSON object which fills in the output fields with the correct values.
It is crucial to critically evaluate the information provided in these responses, recognizing that some of it may be biased, incorrect, or contain duplicates.

You will be provided with a description of each input field and each output field. All of the fields in the output JSON object can be derived using information from the model responses.

{output_format_instruction} Finish your response with a newline character followed by ---

An example is shown below:
---
CHUNK 1 OUTPUT: the text mentions the scientists "Augusta Ada King, Countess of Lovelace" and "Charles Babbage". It states that King was an English mathematician who worked on Babbage's Analytical Engine.

CHUNK 2 OUTPUT: the text passage mentions the scientist "Charles Babbage", who was a mathematician. Therefore, the name output should be ["Charles Babbage"] and the field_of_study output should be ["Mathematician"].

INPUT FIELDS:
- text: a text passage describing scientists

OUTPUT FIELDS:
- name: the list of names for each scientist mentioned in the text
- field_of_study: a list with the field of study for each scientist

Let's think step-by-step in order to answer the question.

REASONING: Looking at both chunk outputs, they specify that the scientists' formal names are "Augusta Ada King" and "Charles Babbage". Chunk Output 2 indicates that Charles Babbage was a Mathematician and Chunk Output 1 says that Augusta Ada King was an English mathematician. Therefore, the name output should be ["Augusta Ada King", "Charles Babbage"] and the field_of_study output should be ["Mathematician", "Mathematician"].

ANSWER:
{{
  "name": ["Augusta Ada King", "Charles Babbage"],
  "field_of_study": ["Mathematician", "Mathematician"]
}}
---
"""

### USER / INSTANCE-SPECIFIC PROMPTS ###
COT_SPLIT_MERGER_BASE_USER_PROMPT = """You are a helpful assistant whose job is to generate a JSON object.
You will be presented with one or more outputs produced by a set of models. Your task is to synthesize these responses into a single, high-quality JSON object which fills in the output fields with the correct values.
It is crucial to critically evaluate the information provided in these responses, recognizing that some of it may be biased, incorrect, or contain duplicates.

You will be provided with a description of each input field and each output field. All of the fields in the output JSON object can be derived using information from the model responses.

{output_format_instruction} Finish your response with a newline character followed by ---
---
{chunk_outputs}

INPUT FIELDS:
{input_fields_desc}

OUTPUT FIELDS:
{output_fields_desc}

Let's think step-by-step in order to answer the question.

REASONING: """



"""This file contains prompts for filter operations."""

### BASE PROMPTS ###
COT_BOOL_BASE_SYSTEM_PROMPT = """You are a helpful assistant whose job is to {job_instruction}.
You will be presented with a context and a filter condition. Output TRUE if the context satisfies the filter condition, and FALSE otherwise.

Remember, your answer must be TRUE or FALSE. Finish your response with a newline character followed by ---

An example is shown below:
---
INPUT FIELDS:
{example_input_fields}

CONTEXT:
{example_context}
{image_disclaimer}
FILTER CONDITION: {example_filter_condition}

Let's think step-by-step in order to answer the question.

REASONING: {example_reasoning}

ANSWER: TRUE
---
"""

COT_BOOL_BASE_USER_PROMPT = """You are a helpful assistant whose job is to {job_instruction}.
You will be presented with a context and a filter condition. Output TRUE if the context satisfies the filter condition, and FALSE otherwise.

Remember, your answer must be TRUE or FALSE. Finish your response with a newline character followed by ---
---
INPUT FIELDS:
{input_fields_desc}

CONTEXT:
{context}
<<image-placeholder>>
FILTER CONDITION: {filter_condition}

Let's think step-by-step in order to answer the question.

REASONING: """


### TEMPLATE INPUTS ###
COT_BOOL_JOB_INSTRUCTION = """answer a TRUE / FALSE question"""
COT_BOOL_IMAGE_JOB_INSTRUCTION = """analyze input image(s) and/or text in order to answer a TRUE / FALSE question"""

COT_BOOL_EXAMPLE_INPUT_FIELDS = """- text: a short passage of text"""
COT_BOOL_IMAGE_EXAMPLE_INPUT_FIELDS = """- image: an image of a scene
- photographer: the photographer of the image"""

COT_BOOL_EXAMPLE_CONTEXT = """{{
  "text": "The quick brown fox jumps over the lazy dog."
}}"""
COT_BOOL_IMAGE_EXAMPLE_CONTEXT = """{{
  "image": <bytes>,
  "photographer": "CameraEnthusiast1"
}}"""

COT_BOOL_EXAMPLE_FILTER_CONDITION = "the text mentions an animal"
COT_BOOL_IMAGE_EXAMPLE_FILTER_CONDITION = "there's an animal in this image"

COT_BOOL_IMAGE_DISCLAIMER = """
<image content provided here; assume in this example the image shows a dog and a cat playing>
"""

COT_BOOL_EXAMPLE_REASONING = """the text mentions the words "fox" and "dog" which are animals, therefore the answer is TRUE."""
COT_BOOL_IMAGE_EXAMPLE_REASONING = """the image shows a dog and a cat playing, both of which are animals, therefore the answer is TRUE."""



from enum import Enum

from palimpzest.query.execution.all_sample_execution_strategy import AllSamplingExecutionStrategy
from palimpzest.query.execution.mab_execution_strategy import MABExecutionStrategy
from palimpzest.query.execution.parallel_execution_strategy import (
    ParallelExecutionStrategy,
    SequentialParallelExecutionStrategy,
)
from palimpzest.query.execution.random_sampling_execution_strategy import RandomSamplingExecutionStrategy
from palimpzest.query.execution.single_threaded_execution_strategy import (
    PipelinedSingleThreadExecutionStrategy,
    SequentialSingleThreadExecutionStrategy,
)


class ExecutionStrategyType(Enum):
    """Available execution strategy types"""
    SEQUENTIAL = SequentialSingleThreadExecutionStrategy
    PIPELINED = PipelinedSingleThreadExecutionStrategy
    PARALLEL = ParallelExecutionStrategy
    SEQUENTIAL_PARALLEL = SequentialParallelExecutionStrategy

class SentinelExecutionStrategyType(Enum):
    MAB = MABExecutionStrategy
    RANDOM = RandomSamplingExecutionStrategy
    ALL = AllSamplingExecutionStrategy



from __future__ import annotations

import time

from palimpzest.constants import NAIVE_EST_NUM_GROUPS, AggFunc
from palimpzest.core.data.dataclasses import OperatorCostEstimates, RecordOpStats
from palimpzest.core.elements.groupbysig import GroupBySig
from palimpzest.core.elements.records import DataRecord, DataRecordSet
from palimpzest.core.lib.schemas import Number
from palimpzest.query.operators.physical import PhysicalOperator


class AggregateOp(PhysicalOperator):
    """
    Aggregate operators accept a list of candidate DataRecords as input to their
    __call__ methods. Thus, we use a slightly modified abstract base class for
    these operators.
    """
    def __call__(self, candidates: DataRecordSet) -> DataRecordSet:
        raise NotImplementedError("Using __call__ from abstract method")


class ApplyGroupByOp(AggregateOp):
    """
    Implementation of a GroupBy operator. This operator groups records by a set of fields
    and applies a function to each group. The group_by_sig object contains the fields to
    group by and the aggregation functions to apply to each group.
    """
    def __init__(self, group_by_sig: GroupBySig, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.group_by_sig = group_by_sig

    def __str__(self):
        op = super().__str__()
        op += f"    Group-by Signature: {str(self.group_by_sig)}\n"
        return op

    def get_id_params(self):
        id_params = super().get_id_params()
        return {"group_by_sig": str(self.group_by_sig.serialize()), **id_params}

    def get_op_params(self):
        op_params = super().get_op_params()
        return {"group_by_sig": self.group_by_sig, **op_params}

    def naive_cost_estimates(self, source_op_cost_estimates: OperatorCostEstimates) -> OperatorCostEstimates:
        # for now, assume applying the groupby takes negligible additional time (and no cost in USD)
        return OperatorCostEstimates(
            cardinality=NAIVE_EST_NUM_GROUPS,
            time_per_record=0,
            cost_per_record=0,
            quality=1.0,
        )

    @staticmethod
    def agg_init(func):
        if func.lower() == "count":
            return 0
        elif func.lower() == "average":
            return (0, 0)
        else:
            raise Exception("Unknown agg function " + func)

    @staticmethod
    def agg_merge(func, state, val):
        if func.lower() == "count":
            return state + 1
        elif func.lower() == "average":
            sum, cnt = state
            return (sum + val, cnt + 1)
        else:
            raise Exception("Unknown agg function " + func)

    @staticmethod
    def agg_final(func, state):
        if func.lower() == "count":
            return state
        elif func.lower() == "average":
            sum, cnt = state
            return float(sum) / cnt
        else:
            raise Exception("Unknown agg function " + func)

    def __call__(self, candidates: DataRecordSet) -> DataRecordSet:
        start_time = time.time()

        # build group array
        agg_state = {}
        for candidate in candidates:
            group = ()
            for f in self.group_by_sig.group_by_fields:
                if not hasattr(candidate, f):
                    raise TypeError(f"ApplyGroupByOp record missing expected field {f}")
                group = group + (getattr(candidate, f),)
            if group in agg_state:
                state = agg_state[group]
            else:
                state = []
                for fun in self.group_by_sig.agg_funcs:
                    state.append(ApplyGroupByOp.agg_init(fun))
            for i in range(0, len(self.group_by_sig.agg_funcs)):
                fun = self.group_by_sig.agg_funcs[i]
                if not hasattr(candidate, self.group_by_sig.agg_fields[i]):
                    raise TypeError(f"ApplyGroupByOp record missing expected field {self.group_by_sig.agg_fields[i]}")
                field = getattr(candidate, self.group_by_sig.agg_fields[i])
                state[i] = ApplyGroupByOp.agg_merge(fun, state[i], field)
            agg_state[group] = state

        # return list of data records (one per group)
        drs = []
        group_by_fields = self.group_by_sig.group_by_fields
        agg_fields = self.group_by_sig.get_agg_field_names()
        for g in agg_state:
            dr = DataRecord(self.group_by_sig.output_schema())
            # NOTE: this will set the parent_id and source_idx to be the id of the final source record;
            #       in the near future we may want to have parent_id accept a list of ids
            dr = DataRecord.from_parent(
                schema=self.group_by_sig.output_schema(),
                parent_record=candidates[-1],
                project_cols=[],
            )
            for i in range(0, len(g)):
                k = g[i]
                setattr(dr, group_by_fields[i], k)
            vals = agg_state[g]
            for i in range(0, len(vals)):
                v = ApplyGroupByOp.agg_final(self.group_by_sig.agg_funcs[i], vals[i])
                setattr(dr, agg_fields[i], v)

            drs.append(dr)

        # create RecordOpStats objects
        total_time = time.time() - start_time
        record_op_stats_lst = []
        for dr in drs:
            record_op_stats = RecordOpStats(
                record_id=dr.id,
                record_parent_id=dr.parent_id,
                record_source_idx=dr.source_idx,
                record_state=dr.to_dict(include_bytes=False),
                full_op_id=self.get_full_op_id(),
                logical_op_id=self.logical_op_id,
                op_name=self.op_name(),
                time_per_record=total_time / len(drs),
                cost_per_record=0.0,
                op_details={k: str(v) for k, v in self.get_id_params().items()},
            )
            record_op_stats_lst.append(record_op_stats)

        # construct and return DataRecordSet
        return DataRecordSet(drs, record_op_stats_lst)


class AverageAggregateOp(AggregateOp):
    # NOTE: we don't actually need / use agg_func here (yet)

    def __init__(self, agg_func: AggFunc, *args, **kwargs):
        kwargs["output_schema"] = Number
        super().__init__(*args, **kwargs)
        self.agg_func = agg_func

        if not self.input_schema.get_desc() == Number.get_desc():
            raise Exception("Aggregate function AVERAGE is only defined over Numbers")

    def __str__(self):
        op = super().__str__()
        op += f"    Function: {str(self.agg_func)}\n"
        return op

    def get_id_params(self):
        id_params = super().get_id_params()
        return {"agg_func": str(self.agg_func), **id_params}

    def get_op_params(self):
        op_params = super().get_op_params()
        return {"agg_func": self.agg_func, **op_params}

    def naive_cost_estimates(self, source_op_cost_estimates: OperatorCostEstimates) -> OperatorCostEstimates:
        # for now, assume applying the aggregation takes negligible additional time (and no cost in USD)
        return OperatorCostEstimates(
            cardinality=1,
            time_per_record=0,
            cost_per_record=0,
            quality=1.0,
        )

    def __call__(self, candidates: DataRecordSet) -> DataRecordSet:
        start_time = time.time()

        # NOTE: this will set the parent_id and source_idx to be the id of the final source record;
        #       in the near future we may want to have parent_id accept a list of ids
        dr = DataRecord.from_parent(schema=Number, parent_record=candidates[-1], project_cols=[])
        dr.value = sum(list(map(lambda c: float(c.value), candidates))) / len(candidates)

        # create RecordOpStats object
        record_op_stats = RecordOpStats(
            record_id=dr.id,
            record_parent_id=dr.parent_id,
            record_source_idx=dr.source_idx,
            record_state=dr.to_dict(include_bytes=False),
            full_op_id=self.get_full_op_id(),
            logical_op_id=self.logical_op_id,
            op_name=self.op_name(),
            time_per_record=time.time() - start_time,
            cost_per_record=0.0,
        )

        return DataRecordSet([dr], [record_op_stats])


class CountAggregateOp(AggregateOp):
    # NOTE: we don't actually need / use agg_func here (yet)

    def __init__(self, agg_func: AggFunc, *args, **kwargs):
        kwargs["output_schema"] = Number
        super().__init__(*args, **kwargs)
        self.agg_func = agg_func

    def __str__(self):
        op = super().__str__()
        op += f"    Function: {str(self.agg_func)}\n"
        return op

    def get_id_params(self):
        id_params = super().get_id_params()
        return {"agg_func": str(self.agg_func), **id_params}

    def get_op_params(self):
        op_params = super().get_op_params()
        return {"agg_func": self.agg_func, **op_params}

    def naive_cost_estimates(self, source_op_cost_estimates: OperatorCostEstimates) -> OperatorCostEstimates:
        # for now, assume applying the aggregation takes negligible additional time (and no cost in USD)
        return OperatorCostEstimates(
            cardinality=1,
            time_per_record=0,
            cost_per_record=0,
            quality=1.0,
        )

    def __call__(self, candidates: DataRecordSet) -> DataRecordSet:
        start_time = time.time()

        # NOTE: this will set the parent_id to be the id of the final source record;
        #       in the near future we may want to have parent_id accept a list of ids
        dr = DataRecord.from_parent(schema=Number, parent_record=candidates[-1], project_cols=[])
        dr.value = len(candidates)

        # create RecordOpStats object
        record_op_stats = RecordOpStats(
            record_id=dr.id,
            record_parent_id=dr.parent_id,
            record_source_idx=dr.source_idx,
            record_state=dr.to_dict(include_bytes=False),
            full_op_id=self.get_full_op_id(),
            logical_op_id=self.logical_op_id,
            op_name=self.op_name(),
            time_per_record=time.time() - start_time,
            cost_per_record=0.0,
            op_details={k: str(v) for k, v in self.get_id_params().items()},
        )

        return DataRecordSet([dr], [record_op_stats])



from __future__ import annotations

import base64
import os
from abc import ABC, abstractmethod
from io import BytesIO

import pandas as pd
from bs4 import BeautifulSoup

from palimpzest import constants
from palimpzest.core.lib.schemas import (
    DefaultSchema,
    File,
    ImageFile,
    PDFFile,
    Schema,
    TextFile,
    WebPage,
    XLSFile,
)
from palimpzest.tools.pdfparser import get_text_from_pdf


# First level of abstraction
class DataReader(ABC):
    """
    The `DataReader` is a base class for which may be used to generate data that
    is processed by PZ.

    Subclasses of the (abstract) `DataReader` class must implement two methods:

    - `__len__()`: which returns the number of elements in the data source
    - `__getitem__(idx: int)`: which takes in an `idx` and returns the element at that index
    """

    def __init__(self, schema: type[Schema] | list[dict]) -> None:
        """
            Constructor for the `DataReader` class.

            Args:
                schema (Schema | list[dict]): The output schema of the records returned by the DataReader
        """
        # NOTE: _schema attribute currently has to match attribute name in Dataset
        self._schema = Schema.from_json(schema) if isinstance(schema, list) else schema

    def __eq__(self, __value: object) -> bool:
        return self.__dict__ == __value.__dict__

    def __str__(self) -> str:
        return f"{self.__class__.__name__}(schema={self.schema})"

    @property
    def schema(self) -> Schema:
        return self._schema

    # NOTE: currently used by optimizer to compute node id for DataReaders
    def serialize(self) -> dict:
        return {"schema": self._schema.json_schema()}

    @abstractmethod
    def __len__(self) -> int:
        """Returns the number of items in the data reader."""
        pass

    @abstractmethod
    def __getitem__(self, idx: int) -> dict:
        """
        Returns a single item from the data reader at the given index.

        Args:
            idx (int): The index of the item to return

        Returns:
            dict: A dictionary representing the item at the given index. The dictionary
                  keys (i.e. fields) should match the fields specified in the schema of the
                  data source, and the values should be the values associated with those fields.

                    # Example return value
                    {"field1": value1, "field2": value2, ...}

        """
        pass


# Second level of abstraction
class DirectoryReader(DataReader):
    """
    DirectoryReader returns a dictionary for each file in a directory. Each dictionary contains the filename and
    contents of a single file in the directory.
    """

    def __init__(self, path: str, schema: Schema) -> None:
        """
        Constructor for the `DirectoryReader` class.

        Args:
            path (str): The path to the directory
            schema (Schema): The output schema of the data source
        """
        assert os.path.isdir(path), f"Path {path} is not a directory"

        self.filepaths = [
            os.path.join(path, filename)
            for filename in sorted(os.listdir(path))
            if os.path.isfile(os.path.join(path, filename))
        ]
        self.path = path
        super().__init__(schema)

    def serialize(self) -> dict:
        return {
            "schema": self.schema.json_schema(),
            "path": self.path,
            "source_type": "directory",
        }

    def __len__(self) -> int:
        return len(self.filepaths)


class FileReader(DataReader):
    """FileReader returns a single dictionary with the filename and contents of a local file (in bytes)."""

    def __init__(self, path: str) -> None:
        """
        Constructor for the `FileReader` class. The `schema` is set to the default `File` schema.

        Args:
            path (str): The path to the file
        """
        super().__init__(File)
        self.filepath = path

    def serialize(self) -> dict:
        return {
            "schema": self.schema.json_schema(),
            "path": self.filepath,
            "source_type": "file",
        }

    def __len__(self) -> int:
        return 1

    def __getitem__(self, idx: int) -> dict:
        """
        Returns a dictionary with the filename and contents of the file.

        Args:
            idx (int): The index of the item to return. This argument is ignored.
        
        Returns:
            dict: A dictionary with the filename and contents of the file.

            .. code-block:: python

                {
                    "filename": "path/to/file.txt",
                    "contents": b"file contents here",
                }
        """
        filename = self.filepath
        with open(self.filepath, "rb") as f:
            contents = f.read()

        return {"filename": filename, "contents": contents}


class MemoryReader(DataReader):
    """
    MemoryReader returns one or more dictionaries that reflect the contents of an in-memory Python object `vals`.
    If `vals` is not a pd.DataFrame, then the dictionary returned by `__getitem__()` has a single field called "value".
    Otherwise, the dictionary contains the key-value mapping from columns to values for the `idx` row in the dataframe.

    TODO(gerardo): Add support for other types of in-memory data structures (he has some code for subclassing
        MemoryReader on his branch)
    """

    def __init__(self, vals: list | pd.DataFrame) -> None:
        """
        Constructor for the `MemoryReader` class. The `schema` is set to the default `DefaultSchema` schema.
        If `vals` is a pd.DataFrame, then the schema is set to the schema inferred from the DataFrame.

        Args:
            vals (Any): The in-memory object to use as the data source
        """
        # if list[dict] --> convert to pd.DataFrame first
        self.vals = pd.DataFrame(vals) if isinstance(vals, list) and all([isinstance(item, dict) for item in vals]) else vals
        schema = Schema.from_df(self.vals) if isinstance(self.vals, pd.DataFrame) else DefaultSchema
        super().__init__(schema)

    def __len__(self) -> int:
        return len(self.vals)

    def __getitem__(self, idx: int) -> dict:
        """
        Returns a dictionary with the value(s) for the element at the specified `idx` in `vals`.

        Args:
            idx (int): The index of the item to return

        Returns:
            dict: If `vals` is not a pd.DataFrame, then the dictionary has a single field called "value".
                Otherwise, the dictionary contains the key-value mapping from columns to values for the
                `idx` row in the dataframe.
            
            .. code-block:: python
            
                # Example return value at idx = 0, for the following list of values
                # [42, 43, 44, ...]
                {"value": 42}

                # Example return value at idx = 0, for the following DataFrame:
                # +---------+---------+---------+
                # |  name   |   job   |  hobby  |
                # +---------+---------+---------+
                # |  Alice  |  doctor |  tennis |
                # |  Bob    |  lawyer |  chess  |
                # +---------+---------+---------+
                {"name": "Alice", "job": "doctor", "hobby": "tennis"}
        """
        item = (
            self.vals.iloc[idx].to_dict()
            if isinstance(self.vals, pd.DataFrame)
            else {"value": self.vals[idx]}
        )

        return item


# Third level of abstraction
class HTMLFileDirectoryReader(DirectoryReader):
    """
    HTMLFileDirectoryReader returns a dictionary for each HTML file in a directory. Each dictionary contains the
    filename, raw HTML content, and parsed content of a single HTML file in the directory.
    """
    def __init__(self, path: str) -> None:
        """
        Constructor for the `HTMLFileDirectoryReader` class. The `schema` is set to the `WebPage` schema.

        Args:
            path (str): The path to the directory
        """
        super().__init__(path=path, schema=WebPage)
        assert all([filename.endswith(tuple(constants.HTML_EXTENSIONS)) for filename in self.filepaths])

    def _html_to_text_with_links(self, html: str) -> str:
        # Parse the HTML content
        soup = BeautifulSoup(html, "html.parser")

        # Find all hyperlink tags
        for a in soup.find_all("a"):
            # Check if the hyperlink tag has an 'href' attribute
            if a.has_attr("href"):
                # Replace the hyperlink with its text and URL in parentheses
                a.replace_with(f"{a.text} ({a['href']})")

        # Extract text from the modified HTML
        text = soup.get_text(separator="\n", strip=True)

        return text

    def __getitem__(self, idx: int) -> dict:
        """
        Returns a dictionary with the filename, raw HTML content, and parsed content of the HTML file at the
        specified `idx`.

        Args:
            idx (int): The index of the item to return
        
        Returns:
            dict: A dictionary with the filename, raw HTML content, and parsed content of the HTML file.

            .. code-block:: python

                {
                    "filename": "file.html",
                    "html": "raw HTML content here",
                    "text": "parsed text content here",
                }
        """
        item = {}
        filepath = self.filepaths[idx]
        item["filename"] = os.path.basename(filepath)
        with open(filepath) as f:
            text_content = f.read()

        html = text_content
        tokens = html.split()[: constants.MAX_HTML_ROWS]
        item["html"] = " ".join(tokens)

        stripped_html = self._html_to_text_with_links(text_content)
        tokens = stripped_html.split()[: constants.MAX_HTML_ROWS]
        item["text"] = " ".join(tokens)

        return item


class ImageFileDirectoryReader(DirectoryReader):
    """
    ImageFileDirectoryReader returns a dictionary for each image file in a directory. Each dictionary contains the
    filename and the base64 encoded bytes content of a single image file in the directory.
    """
    def __init__(self, path: str) -> None:
        """
        Constructor for the `ImageFileDirectoryReader` class. The `schema` is set to the `ImageFile` schema.

        Args:
            path (str): The path to the directory
        """
        super().__init__(path=path, schema=ImageFile)
        assert all([filename.endswith(tuple(constants.IMAGE_EXTENSIONS)) for filename in self.filepaths])

    def __getitem__(self, idx: int) -> dict:
        """
        Returns a dictionary with the filename and base64 encoded bytes content of the image file at the
        specified `idx`.

        Args:
            idx (int): The index of the item to return

        Returns:
            dict: A dictionary with the filename and base64 encoded bytes content of the image file.

            .. code-block:: python

                {
                    "filename": "image.jpg",
                    "contents": b"base64 encoded image content here",
                }
        """
        filepath = self.filepaths[idx]
        filename = os.path.basename(filepath)
        with open(filepath, "rb") as f:
            contents = base64.b64encode(f.read())

        return {"filename": filename, "contents": contents}


class PDFFileDirectoryReader(DirectoryReader):
    """
    PDFFileDirectoryReader returns a dictionary for each PDF file in a directory. Each dictionary contains the
    filename, raw PDF content, and parsed text content of a single PDF file in the directory.

    This class also uses one of a predefined set of PDF processors to extract text content from the PDF files.
    """
    def __init__(
        self,
        path: str,
        pdfprocessor: str = "pypdf",
        file_cache_dir: str = "/tmp",
    ) -> None:
        """
        Constructor for the `PDFFileDirectoryReader` class. The `schema` is set to the `PDFFile` schema.

        Args:
            path (str): The path to the directory
            pdfprocessor (str): The PDF processor to use for extracting text content from the PDF files
            file_cache_dir (str): The directory to store the temporary files generated during PDF processing
        """
        super().__init__(path=path, schema=PDFFile)
        assert all([filename.endswith(tuple(constants.PDF_EXTENSIONS)) for filename in self.filepaths])
        self.pdfprocessor = pdfprocessor
        self.file_cache_dir = file_cache_dir

    def __getitem__(self, idx: int) -> dict:
        """
        Returns a dictionary with the filename, raw PDF content, and parsed text content of the PDF file at the
        specified `idx`.

        Args:
            idx (int): The index of the item to return

        Returns:
            dict: A dictionary with the filename, raw PDF content, and parsed text content of the PDF file.

            .. code-block:: python

                {
                    "filename": "file.pdf",
                    "contents": b"raw PDF content here",
                    "text_contents": "parsed text content here",
                }
        """
        filepath = self.filepaths[idx]
        pdf_filename = os.path.basename(filepath)
        with open(filepath, "rb") as f:
            pdf_bytes = f.read()

        # generate text_content from PDF
        text_content = get_text_from_pdf(pdf_filename, pdf_bytes, pdfprocessor=self.pdfprocessor, file_cache_dir=self.file_cache_dir)

        # construct and return item
        return {"filename": pdf_filename, "contents": pdf_bytes, "text_contents": text_content}


class TextFileDirectoryReader(DirectoryReader):
    """
    TextFileDirectoryReader returns a dictionary for each text file in a directory. Each dictionary contains the
    filename and contents of a single text file in the directory.
    """
    def __init__(self, path: str) -> None:
        """
        Constructor for the `TextFileDirectoryReader` class. The `schema` is set to the `TextFile` schema.

        Args:
            path (str): The path to the directory
        """
        super().__init__(path=path, schema=TextFile)

    def __getitem__(self, idx: int) -> dict:
        """
        Returns a dictionary with the filename and contents of the text file at the specified `idx`.

        Args:
            idx (int): The index of the item to return

        Returns:
            dict: A dictionary with the filename and contents of the text file.

            .. code-block:: python

                {
                    "filename": "file.txt",
                    "contents": "text content here",
                }
        """
        filepath = self.filepaths[idx]
        filename = os.path.basename(filepath)
        with open(filepath) as f:
            contents = f.read()

        return {"filename": filename, "contents": contents}


class XLSFileDirectoryReader(DirectoryReader):
    """
    XLSFileDirectoryReader returns a dictionary for each XLS file in a directory. Each dictionary contains the
    filename, contents, sheet names, and the number of sheets for a single XLS file in the directory.
    """
    def __init__(self, path: str) -> None:
        """
        Constructor for the `XLSFileDirectoryReader` class. The `schema` is set to the `XLSFile` schema.
        """
        super().__init__(path=path, schema=XLSFile)
        assert all([filename.endswith(tuple(constants.XLS_EXTENSIONS)) for filename in self.filepaths])

    def __getitem__(self, idx: int) -> dict:
        """
        Returns a dictionary with the filename, contents, sheet names, and the number of sheets of the XLS file at the
        specified `idx`.

        Args:
            idx (int): The index of the item to return

        Returns:
            dict: A dictionary with the filename, contents, sheet names, and the number of sheets of the XLS file.

            .. code-block:: python

                {
                    "filename": "file.xls",
                    "contents": b"raw XLS content here",
                    "sheet_names": ["Sheet1", "Sheet2", "Sheet3],
                    "number_sheets": 3,
                }
        """
        filepath = self.filepaths[idx]
        filename = os.path.basename(filepath)
        with open(filepath, "rb") as f:
            contents = f.read()

        xls = pd.ExcelFile(BytesIO(contents), engine="openpyxl")

        return {
            "filename": filename,
            "contents": contents,
            "sheet_names": xls.sheet_names,
            "number_sheets": len(xls.sheet_names),
        }



#####################################################
#
#####################################################
# Description: This file contains the functions that are from ASKEM skema tools at endpoints:
# https://api.askem.lum.ai/docs

import requests


def equations_to_latex(image_content):
    url = "https://api.askem.lum.ai/workflows/images/equations-to-latex"
    files = {
        "data": image_content,
    }
    r = requests.post(url, files=files)
    return r.text


def equations_to_latex_base64(image_content):
    url = "https://api.askem.lum.ai/workflows/images/base64/equations-to-latex"
    r = requests.post(url, data=image_content)
    return r.text



from palimpzest.prompts.code_synthesis_prompts import ADVICEGEN_PROMPT, CODEGEN_PROMPT, EXAMPLE_PROMPT
from palimpzest.prompts.prompt_factory import PromptFactory
from palimpzest.prompts.util_phrases import (
    COT_ANSWER_INSTRUCTION,
    COT_REASONING_INSTRUCTION,
    ONE_TO_MANY_OUTPUT_FORMAT_INSTRUCTION,
    ONE_TO_ONE_OUTPUT_FORMAT_INSTRUCTION,
)

__all__ = [
    # code synthesis prompts
    "ADVICEGEN_PROMPT",
    "CODEGEN_PROMPT",
    "EXAMPLE_PROMPT",
    # prompt factory
    "PromptFactory",
    # util phrases
    "COT_ANSWER_INSTRUCTION",
    "COT_REASONING_INSTRUCTION",
    "ONE_TO_MANY_OUTPUT_FORMAT_INSTRUCTION",
    "ONE_TO_ONE_OUTPUT_FORMAT_INSTRUCTION",
]



from __future__ import annotations

from typing import Any

from palimpzest.constants import MODEL_CARDS, Model, PromptStrategy
from palimpzest.core.data.dataclasses import GenerationStats, OperatorCostEstimates
from palimpzest.core.elements.records import DataRecord
from palimpzest.query.generators.generators import generator_factory
from palimpzest.query.operators.convert import LLMConvert

# TYPE DEFINITIONS
FieldName = str


class CriticAndRefineConvert(LLMConvert):

    def __init__(
        self,
        critic_model: Model,
        refine_model: Model,
        *args,
        **kwargs,
    ):
        super().__init__(*args, **kwargs)
        self.critic_model = critic_model
        self.refine_model = refine_model
        
        if self.prompt_strategy == PromptStrategy.COT_QA:
            self.critic_prompt_strategy = PromptStrategy.COT_QA_CRITIC
            self.refinement_prompt_strategy = PromptStrategy.COT_QA_REFINE
        elif self.prompt_strategy == PromptStrategy.COT_QA_IMAGE:
            self.critic_prompt_strategy = PromptStrategy.COT_QA_IMAGE_CRITIC
            self.refinement_prompt_strategy = PromptStrategy.COT_QA_IMAGE_REFINE
        else:
            raise ValueError(f"Unsupported prompt strategy: {self.prompt_strategy}")

        # create generators
        self.critic_generator = generator_factory(self.critic_model, self.critic_prompt_strategy, self.cardinality, self.verbose)
        self.refine_generator = generator_factory(self.refine_model, self.refinement_prompt_strategy, self.cardinality, self.verbose)

    def __str__(self):
        op = super().__str__()
        op += f"    Critic Model: {self.critic_model}\n"
        op += f"    Critic Prompt Strategy: {self.critic_prompt_strategy}\n"
        op += f"    Refine Model: {self.refine_model}\n"
        op += f"    Refinement Prompt Strategy: {self.refinement_prompt_strategy}\n"
        return op

    def get_id_params(self):
        id_params = super().get_id_params()
        id_params = {
            "critic_model": self.critic_model.value,
            "refine_model": self.refine_model.value,
            **id_params,
        }

        return id_params

    def get_op_params(self):
        op_params = super().get_op_params()
        op_params = {
            "critic_model": self.critic_model,
            "refine_model": self.refine_model,
            **op_params,
        }

        return op_params

    def naive_cost_estimates(self, source_op_cost_estimates: OperatorCostEstimates) -> OperatorCostEstimates:
        """
        Currently, we are invoking `self.model`, then critiquing its output with `self.critic_model`, and
        finally refining the output with `self.refine_model`. Thus, we roughly expect to incur the cost
        and time of three LLMConverts. In practice, this naive quality estimate will be overwritten by the
        CostModel's estimate once it executes a few instances of the operator.
        """
        # get naive cost estimates for first LLM call and multiply by 3 for now;
        # of course we should sum individual estimates for each model, but this is a rough estimate
        # and in practice we will need to revamp our naive cost estimates in the near future
        naive_op_cost_estimates = 3 * super().naive_cost_estimates(source_op_cost_estimates)

        # for naive setting, estimate quality as quality of refine model
        model_quality = MODEL_CARDS[self.refine_model.value]["overall"] / 100.0
        naive_op_cost_estimates.quality = model_quality
        naive_op_cost_estimates.quality_lower_bound = naive_op_cost_estimates.quality
        naive_op_cost_estimates.quality_upper_bound = naive_op_cost_estimates.quality

        return naive_op_cost_estimates

    def convert(self, candidate: DataRecord, fields: list[str]) -> tuple[dict[FieldName, list[Any]], GenerationStats]:
        # get input fields
        input_fields = self.get_input_fields()

        # NOTE: when I merge in the `abacus` branch, I will want to update this to reflect the changes I made to reasoning extraction
        # execute the initial model
        original_gen_kwargs = {"project_cols": input_fields, "output_schema": self.output_schema}
        field_answers, reasoning, original_gen_stats, original_messages = self.generator(candidate, fields, **original_gen_kwargs)
        original_output = f"REASONING: {reasoning}\nANSWER: {field_answers}\n"

        # execute the critic model
        critic_gen_kwargs = {"original_output": original_output, "original_messages": original_messages, **original_gen_kwargs}
        _, reasoning, critic_gen_stats, _ = self.critic_generator(candidate, fields, json_output=False, **critic_gen_kwargs)
        critique_output = f"CRITIQUE: {reasoning}\n"

        # execute the refinement model
        refine_gen_kwargs = {"critique_output": critique_output, **critic_gen_kwargs}
        field_answers, reasoning, refine_gen_stats, _ = self.refine_generator(candidate, fields, **refine_gen_kwargs)

        # compute the total generation stats
        generation_stats = original_gen_stats + critic_gen_stats + refine_gen_stats

        return field_answers, generation_stats



"""
This file contains the Generator classes and generator factory.
"""

from __future__ import annotations

import logging
import os
import re
import time
import warnings
from abc import ABC, abstractmethod
from collections import Counter
from copy import deepcopy
from typing import Any, Generic, TypeVar

from colorama import Fore, Style
from openai import OpenAI
from openai.types.chat.chat_completion import ChatCompletion
from together import Together
from together.types.chat_completions import ChatCompletionResponse

from palimpzest.constants import (
    MODEL_CARDS,
    APIClient,
    Cardinality,
    Model,
    PromptStrategy,
)
from palimpzest.core.data.dataclasses import GenerationStats
from palimpzest.core.elements.records import DataRecord
from palimpzest.core.lib.fields import Field, ListField
from palimpzest.prompts import PromptFactory
from palimpzest.query.generators.api_client_factory import APIClientFactory
from palimpzest.utils.generation_helpers import get_json_from_answer
from palimpzest.utils.sandbox import API

# DEFINITIONS
GenerationOutput = tuple[dict, str | None, GenerationStats, list[dict]]
ContextType = TypeVar("ContextType")
InputType = TypeVar("InputType")


logger = logging.getLogger(__name__)

def generator_factory(
    model: Model, prompt_strategy: PromptStrategy, cardinality: Cardinality, verbose: bool = False
) -> BaseGenerator:
    """
    Factory function to return the correct generator based on the model, strategy, and cardinality.
    """
    if model.is_openai_model():
        return OpenAIGenerator(model, prompt_strategy, cardinality, verbose)

    elif model.is_together_model():
        return TogetherGenerator(model, prompt_strategy, cardinality, verbose)

    raise Exception(f"Unsupported model: {model}")


def get_api_key(key: str) -> str:
    # get API key from environment or throw an exception if it's not set
    if key not in os.environ:
        raise ValueError("key not found in environment variables")

    return os.environ[key]


# TODO: make sure answer parsing works with custom prompts / parsers (can defer this)
class BaseGenerator(Generic[ContextType, InputType], ABC):
    """
    Abstract base class for Generators.
    """

    def __init__(
        self,
        model: Model,
        prompt_strategy: PromptStrategy,
        cardinality: Cardinality = Cardinality.ONE_TO_ONE,
        verbose: bool = False,
        system_role: str = "system",
    ):
        self.model = model
        self.model_name = model.value
        self.cardinality = cardinality
        self.prompt_strategy = prompt_strategy
        self.verbose = verbose
        self.system_role = system_role
        self.prompt_factory = PromptFactory(prompt_strategy, model, cardinality)

    @abstractmethod
    def _get_client_or_model(self, **kwargs) -> Any:
        """Returns a client (or local model) which can be invoked to perform the generation."""
        pass

    @abstractmethod
    def _generate_completion(self, client_or_model: Any, payload: dict, **kwargs) -> Any:
        """Generates a completion object using the client (or local model)."""
        pass

    @abstractmethod
    def _get_completion_text(self, completion: Any, **kwargs) -> Any:
        """Extract the completion text from the completion object."""
        pass

    @abstractmethod
    def _get_usage(self, completion: Any, **kwargs) -> Any:
        """Extract the usage statistics from the completion object."""
        pass

    @abstractmethod
    def _get_finish_reason(self, completion: Any, **kwargs) -> Any:
        """Extract the finish reason from the completion object."""
        pass

    @abstractmethod
    def _get_answer_log_probs(self, completion: Any, **kwargs) -> Any:
        """Extract the log probabilities from the completion object."""
        pass

    def _generate_payload(self, messages: list[dict], **kwargs) -> dict:
        """
        Generates the payload which will be fed into the client (or local model).

        Each message will be a dictionary with the following format:
        {
            "role": "user" | "system",
            "type": "text" | "image",
            "content": str
        }
        """
        # get basic parameters
        model = self.model_name
        temperature = kwargs.get("temperature", 0.0)

        # construct messages and add system prompt if present
        chat_messages, user_content = [], []
        for message in messages:
            # flush user content into a message and add system message
            if message["role"] == "system":
                if len(user_content) > 0:
                    chat_messages.append({"role": "user", "content": user_content})
                    user_content = []

                chat_messages.append({"role": self.system_role, "content": message["content"]})

            # add user content for text messages
            elif message["role"] == "user" and message["type"] == "text":
                user_content.append({"type": "text", "text": message["content"]})

            # add user content for image messages
            elif message["role"] == "user" and message["type"] == "image":
                user_content.append({"type": "image_url", "image_url": {"url": message["content"]}})

        # flush any remaining user content into a final message
        if len(user_content) > 0:
            chat_messages.append({"role": "user", "content": user_content})

        # construct and return payload
        payload = {
            "model": model,
            "temperature": temperature,
            "messages": chat_messages,
        }

        return payload

    def _parse_reasoning(self, completion_text: str, **kwargs) -> str:
        """Extract the reasoning for the generated output from the completion object."""
        # use a custom reasoning parser if provided
        if kwargs.get("parse_reasoning"):
            parse_reasoning_fn = kwargs.get("parse_reasoning")
            return parse_reasoning_fn(completion_text)

        # if the model followed the default instructions, the completion text will have reasoning
        # before the "ANSWER:"; if this is the case, we simply extract and return that full section
        if "answer" in completion_text.lower():
            regex = re.compile("(.*?)answer:.*", re.IGNORECASE | re.DOTALL)
            matches = regex.findall(completion_text)
            if len(matches) > 0:
                return matches[0].strip()

        # otherwise, return the full completion text
        return completion_text

    def _prepare_field_answers(self, field_answers: dict | list[dict], fields: dict[str, Field]) -> dict[str, list]:
        """
        field_answers is a dictionary mapping fields to their values. For one-to-one converts, wrap each
        answer in a list. For one-to-many converts, invert the list of dictionaries into a dictionary with
        list values.
        """
        # if this is a one-to-one convert, we need to wrap each answer in a list
        if self.cardinality == Cardinality.ONE_TO_ONE:
            field_answers = {field_name: [field_answers[field_name]] for field_name in fields}

        # otherwise, we need to invert the list of dictionaries into a dictionary with list values
        else:
            field_answers_lst: list[dict] = deepcopy(field_answers)

            field_answers = {field_name: [] for field_name in fields}
            for answer_dict in field_answers_lst:
                for field_name in fields:
                    answer = answer_dict.get(field_name, None)
                    field_answers[field_name].append(answer)

        return field_answers

    def _check_convert_answer_text(self, answer_text: str, fields: dict[str, Field], throw_exception: bool=False) -> dict | list[dict] | None:
        """
        Try parsing the answer text into a JSON object. If the parsing fails, return None.
        """
        try:
            # extract json from the answer text
            field_answers = get_json_from_answer(answer_text, self.model, self.cardinality)

            # TODO: wrap non-list outputs in a list if expected output is a list

            # common error: if the output is a singleton list which contains a list, but the expected field type
            # is a list of strings, or a list of floats, i.e. not a list of lists; then extract the inner list
            for field, field_type in fields.items():
                answer = field_answers[field]
                field_type_is_not_list_of_lists = isinstance(field_type, ListField) and not issubclass(field_type.element_type, ListField)
                answer_is_list_of_lists = isinstance(answer, list) and len(answer) == 1 and isinstance(answer[0], list)
                if field_type_is_not_list_of_lists and answer_is_list_of_lists:
                    field_answers[field] = answer[0]

            # prepare the field answers to match the expected output and return
            return self._prepare_field_answers(field_answers, fields)

        except Exception as e:
            if throw_exception:
                raise e

        return None

    def _check_filter_answer_text(self, answer_text: str) -> dict | None:
        """
        Return {"passed_operator": True} if and only if "true" is in the answer text.
        Return {"passed_operator": False} if and only if "false" is in the answer text.
        Otherwise, return None.
        """
        # NOTE: we may be able to eliminate this condition by specifying this JSON output in the prompt;
        # however, that would also need to coincide with a change to allow the parse_answer_fn to set "passed_operator"
        if "true" in answer_text.lower():
            return {"passed_operator": True}
        elif "false" in answer_text.lower():
            return {"passed_operator": False}

        return None

    def _parse_convert_answer(self, completion_text: str, fields: dict[str, Field], json_output: bool) -> dict[str, list]:
        """Extract the answer from the completion object for convert operations."""
        # if the model followed the default instructions, the completion text will place
        # its answer between "ANSWER:" and "---"
        regex = re.compile("answer:(.*?)---", re.IGNORECASE | re.DOTALL)
        matches = regex.findall(completion_text)
        if len(matches) > 0:
            answer_text = matches[0].strip()

            # if we don't expect a JSON output, return the answer text as is
            if not json_output:
                return answer_text

            # otherwise, try to parse the answer text into a JSON object
            field_answers = self._check_convert_answer_text(answer_text, fields)
            if field_answers is not None:
                return field_answers

        # if the first regex didn't find an answer, try taking all the text after "ANSWER:"
        regex = re.compile("answer:(.*)", re.IGNORECASE | re.DOTALL)
        matches = regex.findall(completion_text)
        if len(matches) > 0:
            answer_text = matches[0].strip()

            # if we don't expect a JSON output, return the answer text as is
            if not json_output:
                return answer_text
            
            # otherwise, try to parse the answer text into a JSON object
            field_answers = self._check_convert_answer_text(answer_text, fields)
            if field_answers is not None:
                return field_answers

        # finally, try taking all of the text; for JSON output, throw an exception if parsing fails
        if not json_output:
            return completion_text

        return self._check_convert_answer_text(completion_text, fields, throw_exception=True)

    def _parse_filter_answer(self, completion_text: str) -> dict[str, list]:
        """Extract the answer from the completion object for filter operations."""
        # if the model followed the default instructions, the completion text will place
        # its answer between "ANSWER:" and "---"
        regex = re.compile("answer:(.*?)---", re.IGNORECASE | re.DOTALL)
        matches = regex.findall(completion_text)
        if len(matches) > 0:
            answer_text = matches[0].strip()
            field_answers = self._check_filter_answer_text(answer_text)
            if field_answers is not None:
                return field_answers

        # if the first regex didn't find an answer, try taking all the text after "ANSWER:"
        regex = re.compile("answer:(.*)", re.IGNORECASE | re.DOTALL)
        matches = regex.findall(completion_text)
        if len(matches) > 0:
            answer_text = matches[0].strip()
            field_answers = self._check_filter_answer_text(answer_text)
            if field_answers is not None:
                return field_answers

        # finally, try taking all of the text; throw an exception if this doesn't work
        field_answers = self._check_filter_answer_text(completion_text)
        if field_answers is None:
            raise Exception(f"Could not parse answer from completion text: {completion_text}")

        return field_answers

    def _parse_answer(self, completion_text: str, fields: dict[str, Field] | None, json_output: bool, **kwargs) -> dict[str, list]:
        """Extract the answer from the completion object."""
        # use a custom answer parser if provided
        if kwargs.get("parse_answer"):
            parse_answer_fn = kwargs.get("parse_answer")
            return parse_answer_fn(completion_text)

        # fields should be a dict if a custom answer parser is not provided
        assert isinstance(fields, dict), "Fields must be provided if a custom answer parser is not provided."

        # extract the per-field answers from the completion text
        field_answers = (
            self._parse_filter_answer(completion_text)
            if self.prompt_strategy.is_bool_prompt()
            else self._parse_convert_answer(completion_text, fields, json_output)
        )

        return field_answers

    def __call__(self, candidate: DataRecord, fields: dict[str, Field] | None, json_output: bool=True, **kwargs) -> GenerationOutput:
        """Take the input record (`candidate`), generate the output `fields`, and return the generated output."""
        client = self._get_client_or_model()
        logger.debug(f"Generating for candidate {candidate} with fields {fields}")

        # fields can only be None if the user provides an answer parser
        fields_check = fields is not None or "parse_answer" in kwargs
        assert fields_check, "`fields` must be provided if `parse_answer` function is not provided in kwargs."

        # if the user (or operator) provides a system prompt instead of a prompt, treat this as
        # the prompt and print a warning
        if "system_prompt" in kwargs and "prompt" not in kwargs:
            kwargs["prompt"] = kwargs["system_prompt"]
            kwargs.pop("system_prompt")
            warnings.warn("Provided `system_prompt` without providing `prompt`; setting `prompt` = `system_prompt`.")  # noqa: B028

        # generate a list of messages which can be used to construct a payload
        messages = self.prompt_factory.create_messages(candidate, fields, **kwargs)

        # create the chat payload
        chat_payload = self._generate_payload(messages, **kwargs)

        # generate the text completion
        start_time = time.time()
        completion = None
        try:
            completion = self._generate_completion(client, chat_payload, **kwargs)
            end_time = time.time()
            logger.debug(f"Generated completion in {end_time - start_time:.2f} seconds")
        # if there's an error generating the completion, we have to return an empty answer
        # and can only account for the time spent performing the failed generation
        except Exception:
            # logger.error(f"Error generating completion: {e}")
            field_answers = {field_name: None for field_name in fields}
            reasoning = None
            generation_stats = GenerationStats(
                model_name=self.model_name,
                llm_call_duration_secs=time.time() - start_time,
                total_llm_calls=1,
            )

            return field_answers, reasoning, generation_stats, messages

        # parse usage statistics and create the GenerationStats
        generation_stats = None
        if completion is not None:
            usage = self._get_usage(completion, **kwargs)
            # finish_reason = self._get_finish_reason(completion, **kwargs)
            # answer_log_probs = self._get_answer_log_probs(completion, **kwargs)

            # get cost per input/output token for the model and parse number of input and output tokens
            usd_per_input_token = MODEL_CARDS[self.model_name]["usd_per_input_token"]
            usd_per_output_token = MODEL_CARDS[self.model_name]["usd_per_output_token"]
            input_tokens = usage["input_tokens"]
            output_tokens = usage["output_tokens"]

            generation_stats = GenerationStats(
                model_name=self.model_name,
                llm_call_duration_secs=end_time - start_time,
                fn_call_duration_secs=0.0,
                total_input_tokens=input_tokens,
                total_output_tokens=output_tokens,
                total_input_cost=input_tokens * usd_per_input_token,
                total_output_cost=output_tokens * usd_per_output_token,
                cost_per_record=input_tokens * usd_per_input_token + output_tokens * usd_per_output_token,
                total_llm_calls=1,
                # "system_prompt": system_prompt,
                # "prompt": prompt,
                # "usage": usage,
                # "finish_reason": finish_reason,
                # "answer_log_probs": answer_log_probs,
                # "answer": answer,
            )

        # pretty print prompt + full completion output for debugging
        completion_text = self._get_completion_text(completion, **kwargs)
        prompt = ""
        for message in messages:
            if message["role"] == "user":
                prompt += message["content"] + "\n" if message["type"] == "text" else "<image>\n"
        logger.debug(f"PROMPT:\n{prompt}")
        logger.debug(Fore.GREEN + f"{completion_text}\n" + Style.RESET_ALL)

        # parse reasoning
        reasoning = None
        try:
            reasoning = self._parse_reasoning(completion_text, **kwargs)
        except Exception:
            # logger.error(f"Error parsing reasoning and answers: {e}")
            logger.debug("TODO: undo this")
            pass

        # parse field answers
        field_answers = None if fields is None else {field_name: None for field_name in fields}
        try:
            field_answers = self._parse_answer(completion_text, fields, json_output, **kwargs)
        except Exception as e:
            # logger.error(f"Error parsing answers: {e}")
            os.makedirs("parse-answer-errors", exist_ok=True)
            ts = time.time()
            with open(f"parse-answer-errors/error-{ts}.txt", "w") as f:
                f.write(f"{str(self.model_name)}\n")
                f.write("#####\n")
                f.write(f"{str(self.prompt_strategy)}\n")
                f.write("#####\n")
                f.write(f"{str(completion_text)}\n")
                f.write("#####\n")
                f.write(f"{str(fields)}\n")
                f.write("#####\n")
                f.write(f"{str(e)}\n")

        logger.debug(f"Generated field answers: {field_answers}")
        return field_answers, reasoning, generation_stats, messages


class OpenAIGenerator(BaseGenerator[str | list[str], str]):
    """
    Class for generating text using the OpenAI chat API.
    """

    def __init__(
        self,
        model: Model,
        prompt_strategy: PromptStrategy,
        cardinality: Cardinality = Cardinality.ONE_TO_ONE,
        verbose: bool = False,
    ):
        # assert that model is an OpenAI model
        assert model.is_openai_model()
        super().__init__(model, prompt_strategy, cardinality, verbose, "developer")

    def _get_client_or_model(self, **kwargs) -> OpenAI:
        """Returns a client (or local model) which can be invoked to perform the generation."""
        return APIClientFactory.get_client(APIClient.OPENAI, get_api_key("OPENAI_API_KEY"))

    def _generate_completion(self, client: OpenAI, payload: dict, **kwargs) -> ChatCompletion:
        """Generates a completion object using the client (or local model)."""
        return client.chat.completions.create(**payload)

    def _get_completion_text(self, completion: ChatCompletion, **kwargs) -> str:
        """Extract the completion text from the completion object."""
        return completion.choices[0].message.content

    def _get_usage(self, completion: ChatCompletion, **kwargs) -> dict:
        """Extract the usage statistics from the completion object."""
        return {
            "input_tokens": completion.usage.prompt_tokens,
            "output_tokens": completion.usage.completion_tokens,
        }

    def _get_finish_reason(self, completion: ChatCompletion, **kwargs) -> str:
        """Extract the finish reason from the completion object."""
        return completion.choices[0].finish_reason

    def _get_answer_log_probs(self, completion: ChatCompletion, **kwargs) -> list[float]:
        """Extract the log probabilities from the completion object."""
        return completion.choices[0].logprobs


class TogetherGenerator(BaseGenerator[str | list[str], str]):
    """
    Class for generating text using the Together chat API.
    """

    def __init__(
        self,
        model: Model,
        prompt_strategy: PromptStrategy,
        cardinality: Cardinality = Cardinality.ONE_TO_ONE,
        verbose: bool = False,
    ):
        # assert that model is a model offered by Together
        assert model.is_together_model()
        super().__init__(model, prompt_strategy, cardinality, verbose, "system")

    def _generate_payload(self, messages: list[dict], **kwargs) -> dict:
        """
        Generates the payload which will be fed into the client (or local model).

        Each message will be a dictionary with the following format:
        {
            "role": "user" | "system",
            "type": "text" | "image",
            "content": str
        }

        For LLAMA3, the payload needs to be in a {"role": <role>, "content": <content>} format.
        """
        # for other models, use our standard payload generation
        if not self.model.is_llama_model():
            return super()._generate_payload(messages, **kwargs)

        # get basic parameters
        model = self.model_name
        temperature = kwargs.get("temperature", 0.0)

        # construct messages in simple {"role": <role>, "content": <content>} format
        chat_messages = []
        for message in messages:
            chat_messages.append({"role": message["role"], "content": message["content"]})

        # construct and return payload
        payload = {
            "model": model,
            "temperature": temperature,
            "messages": chat_messages,
        }

        return payload

    def _get_client_or_model(self, **kwargs) -> Together:
        """Returns a client (or local model) which can be invoked to perform the generation."""
        return APIClientFactory.get_client(APIClient.TOGETHER, get_api_key("TOGETHER_API_KEY"))

    def _generate_completion(self, client: Together, payload: dict, **kwargs) -> ChatCompletionResponse:
        """Generates a completion object using the client (or local model)."""
        return client.chat.completions.create(**payload)

    def _get_completion_text(self, completion: ChatCompletionResponse, **kwargs) -> str:
        """Extract the completion text from the completion object."""
        return completion.choices[0].message.content

    def _get_usage(self, completion: ChatCompletionResponse, **kwargs) -> dict:
        """Extract the usage statistics from the completion object."""
        return {
            "input_tokens": completion.usage.prompt_tokens,
            "output_tokens": completion.usage.completion_tokens,
        }

    def _get_finish_reason(self, completion: ChatCompletionResponse, **kwargs) -> str:
        """Extract the finish reason from the completion object."""
        return completion.choices[0].finish_reason.value

    def _get_answer_log_probs(self, completion: ChatCompletionResponse, **kwargs) -> list[float]:
        """Extract the log probabilities from the completion object."""
        return completion.choices[0].logprobs


### CODE SYNTHESIS EXECUTION ###
def code_execution(api: API, code: str, candidate_dict: dict[str, Any], verbose: bool = False):
    inputs = {field_name: candidate_dict[field_name] for field_name in api.inputs}
    response = api.api_execute(code, inputs)
    pred = response["response"] if response["status"] and response["response"] else None
    return pred


def code_ensemble_execution(
    api: API, code_ensemble: dict[str, str], candidate_dict: dict[str, Any], verbose: bool = True
) -> GenerationOutput:
    start_time = time.time()
    try:
        preds = list()
        for _, code in code_ensemble.items():
            pred = code_execution(api, code, candidate_dict)
            preds.append(pred)

        preds = [pred for pred in preds if pred is not None]

        if len(preds) == 1:
            majority_response = preds[0]
            exec_stats = GenerationStats(fn_call_duration_secs=time.time() - start_time)
            return majority_response, None, exec_stats

        if len(preds) > 0:
            majority_response = Counter(preds).most_common(1)[0][0]
            exec_stats = GenerationStats(fn_call_duration_secs=time.time() - start_time)
            return majority_response, None, exec_stats

    except Exception:
        pass

    return None, None, GenerationStats(fn_call_duration_secs=time.time() - start_time)



import logging

import numpy as np

from palimpzest.core.data.dataclasses import SentinelPlanStats
from palimpzest.core.elements.records import DataRecord, DataRecordSet
from palimpzest.query.execution.execution_strategy import SentinelExecutionStrategy
from palimpzest.query.operators.physical import PhysicalOperator
from palimpzest.query.operators.scan import ScanPhysicalOp
from palimpzest.query.optimizer.plan import SentinelPlan
from palimpzest.utils.progress import create_progress_manager

logger = logging.getLogger(__name__)

class OpSet:
    """
    This class represents the set of operators which are currently in the frontier for a given logical operator.
    Each operator in the frontier is an instance of a PhysicalOperator which either:

    1. lies on the Pareto frontier of the set of sampled operators, or
    2. has been sampled fewer than j times
    """

    def __init__(self, op_set: list[PhysicalOperator], source_indices: list[int]):
        # construct the set of operators
        self.ops = op_set

        # store the order in which we will sample the source records
        self.source_indices = source_indices

        # set the initial inputs for this logical operator
        is_scan_op = isinstance(op_set[0], ScanPhysicalOp)
        self.source_idx_to_input = {source_idx: [source_idx] for source_idx in self.source_indices} if is_scan_op else {}

    def get_op_input_pairs(self) -> list[PhysicalOperator, DataRecord | int | None]:
        """
        Returns the list of frontier operators and their next input to process. If there are
        any indices in `source_indices_to_sample` which this operator does not sample on its own, then
        we also have this frontier process that source_idx's input with its max quality operator.
        """
        # get the list of (op, source_idx) pairs which this operator needs to execute
        op_source_idx_pairs = []
        for op in self.ops:
            # construct list of inputs by looking up the input for the given source_idx
            for source_idx in self.source_indices:
                op_source_idx_pairs.append((op, source_idx))

        # fetch the corresponding (op, input) pairs
        op_input_pairs = []
        for op, source_idx in op_source_idx_pairs:
            op_input_pairs.extend([(op, input_record) for input_record in self.source_idx_to_input[source_idx]])

        return op_input_pairs

    def pick_highest_quality_output(self, record_sets: list[DataRecordSet]) -> DataRecordSet:
        # if there's only one operator in the set, we return its record_set
        if len(record_sets) == 1:
            return record_sets[0]

        # NOTE: I don't like that this assumes the models are consistent in
        #       how they order their record outputs for one-to-many converts;
        #       eventually we can try out more robust schemes to account for
        #       differences in ordering
        # aggregate records at each index in the response
        idx_to_records = {}
        for record_set in record_sets:
            for idx in range(len(record_set)):
                record, record_op_stats = record_set[idx], record_set.record_op_stats[idx]
                if idx not in idx_to_records:
                    idx_to_records[idx] = [(record, record_op_stats)]
                else:
                    idx_to_records[idx].append((record, record_op_stats))

        # compute highest quality answer at each index
        out_records = []
        out_record_op_stats = []
        for idx in range(len(idx_to_records)):
            records_lst, record_op_stats_lst = zip(*idx_to_records[idx])
            max_quality_record, max_quality = records_lst[0], record_op_stats_lst[0].quality
            max_quality_stats = record_op_stats_lst[0]
            for record, record_op_stats in zip(records_lst[1:], record_op_stats_lst[1:]):
                record_quality = record_op_stats.quality
                if record_quality > max_quality:
                    max_quality_record = record
                    max_quality = record_quality
                    max_quality_stats = record_op_stats
            out_records.append(max_quality_record)
            out_record_op_stats.append(max_quality_stats)

        # create and return final DataRecordSet
        return DataRecordSet(out_records, out_record_op_stats)

    def update_inputs(self, source_idx_to_record_sets: dict[int, DataRecordSet]):
        """
        Update the inputs for this logical operator based on the outputs of the previous logical operator.
        """
        for source_idx, record_sets in source_idx_to_record_sets.items():
            input = []
            max_quality_record_set = self.pick_highest_quality_output(record_sets)
            for record in max_quality_record_set:
                input.append(record if record.passed_operator else None)

            self.source_idx_to_input[source_idx] = input


class AllSamplingExecutionStrategy(SentinelExecutionStrategy):

    def _get_source_indices(self):
        """Get the list of source indices which the sentinel plan should execute over."""
        # create list of all source indices and shuffle it
        total_num_samples = len(self.val_datasource)
        source_indices = list(np.arange(total_num_samples))

        return source_indices

    def _execute_sentinel_plan(self,
            plan: SentinelPlan,
            op_sets: dict[str, OpSet],
            expected_outputs: dict[int, dict] | None,
            plan_stats: SentinelPlanStats,
        ) -> SentinelPlanStats:
        # execute operator sets in sequence
        for op_idx, (logical_op_id, op_set) in enumerate(plan):
            # get frontier ops and their next input
            op_input_pairs = op_sets[logical_op_id].get_op_input_pairs()

            # break out of the loop if op_input_pairs is empty, as this means all records have been filtered out
            if len(op_input_pairs) == 0:
                break

            # run sampled operators on sampled inputs
            source_idx_to_record_sets_and_ops, _ = self._execute_op_set(op_input_pairs)

            # FUTURE TODO: have this return the highest quality record set simply based on our posterior (or prior) belief on operator quality
            # get the target record set for each source_idx
            source_idx_to_target_record_set = self._get_target_record_sets(logical_op_id, source_idx_to_record_sets_and_ops, expected_outputs)

            # TODO: make consistent across here and RandomSampling
            # FUTURE TODO: move this outside of the loop (i.e. assume we only get quality label(s) after executing full program)
            # score the quality of each generated output
            physical_op_cls = op_set[0].__class__
            source_idx_to_record_sets = {
                source_idx: list(map(lambda tup: tup[0], record_sets_and_ops))
                for source_idx, record_sets_and_ops in source_idx_to_record_sets_and_ops.items()
            }
            source_idx_to_record_sets = self._score_quality(physical_op_cls, source_idx_to_record_sets, source_idx_to_target_record_set)

            # flatten the lists of records and record_op_stats
            all_records, all_record_op_stats = self._flatten_record_sets(source_idx_to_record_sets)

            # update plan stats
            plan_stats.add_record_op_stats(all_record_op_stats)

            # add records (which are not filtered) to the cache, if allowed
            self._add_records_to_cache(logical_op_id, all_records)

            # FUTURE TODO: simply set input based on source_idx_to_target_record_set (b/c we won't have scores computed)
            # provide the champion record sets as inputs to the next logical operator
            if op_idx + 1 < len(plan):
                next_logical_op_id = plan.logical_op_ids[op_idx + 1]
                op_sets[next_logical_op_id].update_inputs(source_idx_to_record_sets)

        # close the cache
        self._close_cache(plan.logical_op_ids)

        # finalize plan stats
        plan_stats.finish()

        return plan_stats

    def execute_sentinel_plan(self, plan: SentinelPlan, expected_outputs: dict[int, dict] | None):
        """
        NOTE: this function currently requires us to set k and j properly in order to make
              comparison in our research against the corresponding sample budget in MAB.

        NOTE: the number of samples will slightly exceed the sample_budget if the number of operator
        calls does not perfectly match the sample_budget. This may cause some minor discrepancies with
        the progress manager as a result.
        """
        # for now, assert that the first operator in the plan is a ScanPhysicalOp
        assert all(isinstance(op, ScanPhysicalOp) for op in plan.operator_sets[0]), "First operator in physical plan must be a ScanPhysicalOp"
        logger.info(f"Executing plan {plan.plan_id} with {self.max_workers} workers")
        logger.info(f"Plan Details: {plan}")

        # initialize plan stats
        plan_stats = SentinelPlanStats.from_plan(plan)
        plan_stats.start()

        # get list of source indices which can be sampled from
        source_indices = self._get_source_indices()

        # initialize set of physical operators for each logical operator
        op_sets = {
            logical_op_id: OpSet(op_set, source_indices)
            for logical_op_id, op_set in plan
        }

        # initialize and start the progress manager
        self.progress_manager = create_progress_manager(plan, sample_budget=self.sample_budget, progress=self.progress)
        self.progress_manager.start()

        # NOTE: we must handle progress manager outside of _exeecute_sentinel_plan to ensure that it is shut down correctly;
        #       if we don't have the `finally:` branch, then program crashes can cause future program runs to fail because
        #       the progress manager cannot get a handle to the console 
        try:
            # execute sentinel plan by sampling records and operators
            plan_stats = self._execute_sentinel_plan(plan, op_sets, expected_outputs, plan_stats)

        finally:
            # finish progress tracking
            self.progress_manager.finish()

        logger.info(f"Done executing sentinel plan: {plan.plan_id}")
        logger.debug(f"Plan stats: (plan_cost={plan_stats.total_plan_cost}, plan_time={plan_stats.total_plan_time})")

        return plan_stats



from __future__ import annotations

import os
import time
from typing import Callable

from chromadb.api.models.Collection import Collection
from chromadb.utils.embedding_functions import SentenceTransformerEmbeddingFunction
from chromadb.utils.embedding_functions.openai_embedding_function import OpenAIEmbeddingFunction
from openai import OpenAI
from sentence_transformers import SentenceTransformer

from palimpzest.constants import MODEL_CARDS, Model
from palimpzest.core.data.dataclasses import GenerationStats, OperatorCostEstimates, RecordOpStats
from palimpzest.core.elements.records import DataRecord, DataRecordSet
from palimpzest.core.lib.schemas import Schema
from palimpzest.query.operators.physical import PhysicalOperator


class RetrieveOp(PhysicalOperator):
    def __init__(
        self,
        index: Collection,
        search_attr: str,
        output_attrs: list[dict] | type[Schema],
        search_func: Callable | None,
        k: int,
        *args,
        **kwargs,
    ) -> None:
        """
        Initialize the RetrieveOp object.
        
        Args:
            index (Collection): The PZ index to use for retrieval.
            search_attr (str): The attribute to search on.
            output_attrs (list[dict]): The output fields containing the results of the search.
            search_func (Callable | None): The function to use for searching the index. If None, the default search function will be used.
            k (int): The number of top results to retrieve.
        """
        super().__init__(*args, **kwargs)

        # extract the field names from the output_attrs
        if isinstance(output_attrs, Schema):
            self.output_field_names = output_attrs.field_names()
        elif isinstance(output_attrs, list):
            self.output_field_names = [attr["name"] for attr in output_attrs]
        else:
            raise ValueError("`output_attrs` must be a list of dicts or a Schema object.")

        if len(self.output_field_names) != 1 and search_func is None:
            raise ValueError("If `search_func` is None, `output_attrs` must have a single field.")

        self.index = index
        self.search_attr = search_attr
        self.output_attrs = output_attrs
        self.search_func = search_func if search_func is not None else self.default_search_func
        self.k = k

    def __str__(self):
        op = super().__str__()
        op += f"    Retrieve: {self.index.__class__.__name__} with top {self.k}\n"
        return op

    def get_id_params(self):
        id_params = super().get_id_params()
        id_params = {
            "index": self.index.__class__.__name__,
            "search_attr": self.search_attr,
            "output_attrs": self.output_attrs,
            "k": self.k,
            **id_params,
        }

        return id_params

    def get_op_params(self):
        op_params = super().get_op_params()
        op_params = {
            "index": self.index,
            "search_func": self.search_func,
            "search_attr": self.search_attr,
            "output_attrs": self.output_attrs,
            "k": self.k,
            **op_params,
        }

        return op_params

    def naive_cost_estimates(self, source_op_cost_estimates: OperatorCostEstimates) -> OperatorCostEstimates:
        """
        Compute naive cost estimates for the Retrieve operation. These estimates assume
        that the Retrieve (1) has no cost and (2) has perfect quality.
        """
        return OperatorCostEstimates(
            cardinality=source_op_cost_estimates.cardinality,
            time_per_record=0.01 * self.k,   # estimate 10 ms execution lookup per output
            cost_per_record=0.001 * self.k,  # estimate small marginal cost of lookups 
            quality=1.0,
        )

    def default_search_func(self, index: Collection, query: list[str] | list[list[float]], k: int) -> list[str] | list[list[str]]:
        """
        Default search function for the Retrieve operation. This function uses the index to
        retrieve the top-k results for the given query. The query will be a (possibly singleton)
        list of strings or a list of lists of floats (i.e., embeddings). The function will return
        the top-k results per-query in (descending) sorted order. If the input is a singleton list,
        then the output will be a list of strings. If the input is a list of lists, then the output
        will be a list of lists of strings.

        Args:
            index (PZIndex): The index to use for retrieval.
            query (list[str] | list[list[float]]): The query (or queries) to search for.
            k (int): The maximum number of results the retrieve operator will return.

        Returns:
            list[str] | list[list[str]]: The top results in (descending) sorted order per query.
        """
        # check if the input is a singleton list or a list of lists
        is_singleton_list = len(query) == 1

        if isinstance(index, Collection):
            # if the index is a chromadb collection, use the query method
            results = index.query(query, n_results=k)

            # the results["documents"] will be a list[list[str]]; if the input is a singleton list,
            # then we output the list of strings (i.e., the first element of the list), otherwise
            # we output the list of lists
            final_results = results["documents"][0] if is_singleton_list else results["documents"]

            # NOTE: self.output_field_names must be a singleton for default_search_func to be used
            return {self.output_field_names[0]: final_results}

        else:
            raise ValueError("Unsupported index type. Must be either a Collection.")

    def _create_record_set(
        self,
        candidate: DataRecord,
        top_k_results: dict[str, list[str] | list[list[str]]] | None,
        generation_stats: GenerationStats,
        total_time: float,
    ) -> DataRecordSet:
        """
        Given an input DataRecord and the top_k_results, construct the resulting RecordSet.
        """
        # create output DataRecord an set the output attribute
        output_dr, answer = DataRecord.from_parent(self.output_schema, parent_record=candidate), {}
        for output_field_name in self.output_field_names:
            top_k_attr_results = None if top_k_results is None else top_k_results[output_field_name]
            setattr(output_dr, output_field_name, top_k_attr_results)
            answer[output_field_name] = top_k_attr_results

        # get the record_state and generated fields
        record_state = output_dr.to_dict(include_bytes=False)

        # NOTE: this should be equivalent to self.get_fields_to_generate()
        generated_fields = self.output_field_names

        # construct the RecordOpStats object
        record_op_stats = RecordOpStats(
            record_id=output_dr.id,
            record_parent_id=output_dr.parent_id,
            record_source_idx=output_dr.source_idx,
            record_state=record_state,
            full_op_id=self.get_full_op_id(),
            logical_op_id=self.logical_op_id,
            op_name=self.op_name(),
            time_per_record=total_time,
            cost_per_record=generation_stats.cost_per_record,
            answer=answer,
            input_fields=self.input_schema.field_names(),
            generated_fields=generated_fields,
            fn_call_duration_secs=total_time - generation_stats.llm_call_duration_secs,
            llm_call_duration_secs=generation_stats.llm_call_duration_secs,
            total_llm_calls=generation_stats.total_llm_calls,
            total_embedding_llm_calls=generation_stats.total_embedding_llm_calls,
            op_details={k: str(v) for k, v in self.get_id_params().items()},
        )

        drs = [output_dr]
        record_op_stats_lst = [record_op_stats]

        # construct and return the record set
        return DataRecordSet(drs, record_op_stats_lst)


    def __call__(self, candidate: DataRecord) -> DataRecordSet:
        start_time = time.time()

        # check that query is a string or list of strings, otherwise return output with self.output_field_names set to None
        query = getattr(candidate, self.search_attr)
        query_is_str = isinstance(query, str)
        query_is_list_of_str = isinstance(query, list) and all(isinstance(q, str) for q in query)
        if not query_is_str and not query_is_list_of_str:
            return self._create_record_set(
                candidate=candidate,
                top_k_results=None,
                generation_stats=GenerationStats(),
                total_time=time.time() - start_time,
            )

        # if query is a string, convert it to a list of strings
        if query_is_str:
            query = [query]

        # compute input/query embedding(s) if the index is a chromadb collection
        inputs, gen_stats = None, GenerationStats()
        if isinstance(self.index, Collection):
            uses_openai_embedding_fcn = isinstance(self.index._embedding_function, OpenAIEmbeddingFunction)
            uses_sentence_transformer_embedding_fcn = isinstance(self.index._embedding_function, SentenceTransformerEmbeddingFunction)
            error_msg = "ChromaDB index must use OpenAI or SentenceTransformer embedding function; see: https://docs.trychroma.com/integrations/embedding-models/openai"
            assert uses_openai_embedding_fcn or uses_sentence_transformer_embedding_fcn, error_msg

            model_name = self.index._embedding_function._model_name if uses_openai_embedding_fcn else "clip-ViT-B-32"
            err_msg = f"For Chromadb, we currently only support `text-embedding-3-small` and `clip-ViT-B-32`; your index uses: {model_name}"
            embedding_model_names = [model.value for model in Model if model.is_embedding_model()]
            assert model_name in embedding_model_names, err_msg

            # compute embeddings
            try:
                embed_start_time = time.time()
                total_input_tokens = 0.0
                if uses_openai_embedding_fcn:
                    client = OpenAI()
                    response = client.embeddings.create(input=query, model=model_name)
                    total_input_tokens = response.usage.total_tokens
                    inputs = [item.embedding for item in response.data]

                elif uses_sentence_transformer_embedding_fcn:
                    model = SentenceTransformer(model_name)
                    inputs = model.encode(query)

                embed_total_time = time.time() - embed_start_time

                # compute cost of embedding(s)
                model_card = MODEL_CARDS[model_name]
                total_input_cost = model_card["usd_per_input_token"] * total_input_tokens
                gen_stats = GenerationStats(
                    model_name=model_name,
                    total_input_tokens=total_input_tokens,
                    total_output_tokens=0.0,
                    total_input_cost=total_input_cost,
                    total_output_cost=0.0,
                    cost_per_record=total_input_cost,
                    llm_call_duration_secs=embed_total_time,
                    total_llm_calls=1,
                    total_embedding_llm_calls=len(query),
                )
            except Exception:
                query = None

        # in the default case, pass string inputs rather than embeddings
        if inputs is None:
            inputs = query

        try:
            assert inputs is not None, "Error: inputs is None (likely because embedding generation failed)"
            top_results = self.search_func(self.index, inputs, self.k)

        except Exception:
            top_results = ["error-in-retrieve"]
            os.makedirs("retrieve-errors", exist_ok=True)
            ts = time.time()
            with open(f"retrieve-errors/error-{ts}.txt", "w") as f:
                f.write(str(query))

        # TODO: the user is always right! let's drop this post-processing in the future
        # filter top_results for the top_k_results
        top_k_results = {output_field_name: [] for output_field_name in self.output_field_names}
        for output_field_name in self.output_field_names:
            if output_field_name in top_results:
                if all([isinstance(result, list) for result in top_results[output_field_name]]):
                    for result in top_results[output_field_name]:
                        top_k_results[output_field_name].append(result[:self.k])
                else:
                    top_k_results[output_field_name] = top_results[output_field_name][:self.k]
            else:
                top_k_results[output_field_name] = []

        if self.verbose:
            print(f"Top {self.k} results: {top_k_results}")

        # construct and return the record set
        return self._create_record_set(
            candidate=candidate,
            top_k_results=top_k_results,
            generation_stats=gen_stats,
            total_time=time.time() - start_time,
        )
